"""Logging submodule for the nonadditivity analysis code."""

import logging
import sys
from collections.abc import Callable
from functools import wraps
from importlib import import_module
from pathlib import Path
from typing import Any

_ = logging.getLogger(__name__)


def setup_logging(
    level: int = logging.INFO,
    out_file: Path | None = None,
) -> None:
    """Set up module wide logging.

    If out_file is given the log is printed to
    stdout as well as into the out_file.

    Args:
        level (int, optional): logging level. Defaults to logging.INFO.
        out_file (Path | None, optional): file to write logs to. Defaults
        to None.
    """
    # set correct log level
    logging.root.setLevel(level)
    logging.root.handlers.clear()

    # create console handler and set level to info
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        fmt="%(asctime)s %(module)s %(levelname)-7s %(message)s",
        datefmt="%Y/%b/%d %H:%M:%S",
    )
    handler.setFormatter(formatter)
    logging.root.addHandler(handler)

    if out_file:
        # create file log
        out_file = out_file.absolute()

        console = logging.FileHandler(out_file, mode="a")
        console.setFormatter(formatter)
        logging.root.addHandler(console)


def get_logger() -> logging.Logger:
    """Get logger for nonadditivity module.

    Returns:
        logging.Logger: setup logger.
    """
    logger = logging.getLogger()
    return logger


def log_versions(
    logger: logging.Logger,
    packages: list[str],
    workflow: str,
) -> Callable:
    """Log package names and versions for decorated workflow.

    Decorator for functions that takes the logger as well as a list of
    strings containing valid package names and prints logs the versions
    of the package names given in packages.

    Args:
        logger (logging.Logger): logger to use
        packages (list[str]): packages to log versions for
        workflow: (str, optional): name of the workflow. Defaults to Free-
        Wilson Analysis.

    Returns:
        Callable: wrapped function
    """

    def inner(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable[..., Any]:
            versions = ", ".join(
                [f"{p} v{import_module(p).__version__}" for p in packages],
            )
            logger.info("Running %s using %s", workflow, versions)
            return func(*args, **kwargs)

        return wrapper

    return inner



"""Nonadditivity Calculation Workflow."""

from nonadditivity import (
    classify,
    parse_input_file,
    run_mmpdlib_code,
    run_nonadditivity_core,
    write_output_files,
    write_smiles_id_file,
)
from nonadditivity.utils.commandline import InputOptions
from nonadditivity.utils.log import get_logger, log_versions

logger = get_logger()


@log_versions(
    logger=logger,
    packages=["mmpdblib", "rdkit", "pandas"],
    workflow="Nonadditivity Analysis",
)
def run_nonadd_calculation(input_options: InputOptions) -> None:
    """Run the Nonadditivity analysis package.

    run_control needs to have the following attributes:
    - infile (required)
    - outfile (optional)
    - property_columns (can be empty)
    - units (as many as property_columns or empty)
    - max_heavy (required)
    - update (Boolean)
    - no_chiral(Boolean)
    - write_images(Boolean)
    """
    verbose = input_options.verbose < 30

    update = input_options.update
    directory = input_options.directory
    series_column = input_options.series_column
    include_censored = input_options.include_censored

    error_path = directory / "problem_smiles.smi"
    structure_filename = directory / "ligands.smi"
    mmp_output_path = directory / "mmp_raw.csv"
    naa_file_path = directory / "NAA_output.csv"
    per_compound_path = directory / "perCompound.csv"
    c2c_file_path = directory / "c2c.csv"
    fragment_file_path = directory / "fragments.json"

    per_compound_dataframe, property_columns = parse_input_file(
        infile=input_options.infile,
        error_path=error_path,
        property_columns=input_options.property_columns,
        max_heavy=input_options.max_heavy,
        units=input_options.units,
        delimiter=input_options.delimiter,
        series_column=series_column,
        verbose=verbose,
    )

    write_smiles_id_file(
        per_compound_dataframe=per_compound_dataframe,
        smifile=structure_filename,
    )
    mmp_dataframe = run_mmpdlib_code(
        cache=update,
        fragments_file=fragment_file_path,
        mmp_outputfile=mmp_output_path,
        max_variable_heavies=input_options.max_heavy_in_transformation,
        structure_file=structure_filename,
    )
    na_dataframe, _, _, c2c_dataframe = run_nonadditivity_core(
        mmp_dataframe=mmp_dataframe,
        per_compound_dataframe=per_compound_dataframe,
        property_columns=property_columns,
        no_chiral=input_options.no_chiral,
        series_column=series_column,
        include_censored=include_censored,
        verbose=verbose,
    )
    if input_options.classify:
        _ = classify(
            per_compound_dataframe=per_compound_dataframe,
            na_dataframe=na_dataframe,
        )

    write_output_files(
        per_compound_dataframe=per_compound_dataframe,
        na_dataframe=na_dataframe,
        c2c_dataframe=c2c_dataframe,
        c2c_file_path=c2c_file_path,
        naa_file_path=naa_file_path,
        per_compound_path=per_compound_path,
        property_columns=property_columns,
        include_censored=include_censored,
        series_column=series_column,
        canonicalize=input_options.canonicalize,
    )



"""Functionality for r group distance calculation.

Contains all the functionality for evalutating the distance between two
r groups of a double transformation cycle over all levels (compound,
transformation and circle). This is part of the classification workflow
of the Nonadditivity Analysis.
"""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from rdkit import Chem
from rdkit.Chem import rdmolfiles

from nonadditivity.classification.circle_classification import get_combinations
from nonadditivity.classification.utils import is_match_unique
from nonadditivity.utils.log import get_logger

if TYPE_CHECKING:
    from nonadditivity.classification import Transformation
    from nonadditivity.utils.types import Molecule

logger = get_logger()


# -------------------------------------------------------------------------
#
#               TRANSFORMATION LEVEL OF THINGS
#
# =========================================================================


def all_matches_symetrically_interchangeable(
    matches: tuple[tuple[int, ...], ...],
    molecule: Molecule,
) -> bool:
    """Check whether multiple matches are symmetyrically interchangable.

    Checks if two substructure matches are symetrically interchangeable.
    E.g., a methyl has two ortho substitued fluorines, which are both matched
    once if you want to match [c1ccccc1F]. Both matches are equivalent.

    Args:
        matches (tuple[tuple[int]]): list of possible matches
        molecule (Molecule): molecule the matches are from.

    Returns:
        bool: True if matches are symetrically interchangeable.
    """
    if len(matches) != 2:
        return False
    difference = sorted(set(matches[0]).symmetric_difference(matches[1]))
    if len(difference) % 2 != 0:
        return False
    ranks = list(
        rdmolfiles.CanonicalRankAtoms(  # pylint: disable=I1101
            molecule,
            breakTies=False,
        ),
    )
    ranks = sorted([ranks[d] for d in difference])
    equal = True
    for index in range(0, len(ranks), 2):
        if ranks[index] == ranks[index + 1]:
            continue
        equal = False
    return equal


def all_matches_chemically_equivalent(
    matches: tuple[tuple[int, ...], ...],
    molecule: Molecule,
) -> bool:
    """Check whether all matches are chemically equivalent.

    Checks wheter two matches of a substructure are chemically
    equivalent, so lets say we try to match C-F and our molecule
    has a CF3 group. This group is matched three times, even tho
    every match is equivalent for our purposes.

    Args:
        matches (tuple[tuple[int]]): Matches
        molecule (Molecule): molecule matches are found

    Returns:
        bool: True if all matches are chemically equivalent.
    """
    differences = set()
    if len(matches) == 2:
        for match in matches:
            differences = differences.symmetric_difference(set(match))
    elif len(matches) == 3:
        diff1 = set(matches[0]).symmetric_difference(set(matches[1]))
        diff2 = set(matches[0]).symmetric_difference(set(matches[2]))
        differences = diff1.union(diff2)
    else:
        return False
    if len(differences) != len(matches):
        return False
    neighbors = set()
    for index in differences:
        for neighbor in molecule.GetAtomWithIdx(index).GetNeighbors():
            neighbors.add(neighbor.GetIdx())
    return len(neighbors) == 1


def get_correct_match(
    single_match: tuple[int, ...],
    multiple_matches: tuple[tuple[int, ...], ...],
    num_attachment_points: int,
) -> list[int] | tuple[int, ...]:
    """Get the correct match.

    You have a list A and a list of lists B and a given length of the
    intersection between the single list and one of the lists in the list of lists
    should have. returns the list from the list of lists that has the right length
    intersection with the list A.
    if no list in B has the right length intersection with A, an empty list is returned

    Args:
        single_match (list[int]): list A
        multiple_matches (tuple[tuple[int]]): list of lists B
        num_attachment_points (int): length of intersection

    Returns:
        list[int]: [] or list out of B with correct length Intersection.
    """
    single_set = set(single_match)
    for match in multiple_matches:
        if len(single_set.intersection(set(match))) == num_attachment_points:
            return match
    return []


def remove_stereo_at_beginning(smarts: str) -> str:
    """Remove stereo information from smarts.

    Removes E/Z information from cut smarts, so that
    /C/CCC becomes CCCC.

    Args:
        smarts (str): smarts to check.

    Returns:
        str: smarts without invalid stereo information
    """
    exchanged = False
    for start in ("\\", "/"):
        if smarts.startswith(start):
            smarts = smarts.lstrip(start)
            exchanged = True
            break
    if not exchanged:
        return smarts
    index = -1
    for ind, char in enumerate(smarts):
        if char in ("/", "\\"):
            index = ind
            break
    if index == -1:
        return smarts
    return smarts[:index] + smarts[index + 1 :]


def get_matches_without_hydrogen(
    constant_smarts: str,
    transformation_smarts: str,
    molecule: Molecule,
) -> tuple[list[int], ...] | tuple[list[list[int]], ...]:
    """Get matches without hydrogen.

    This function tries to uniquely match transformation_smarts and constant_smarts in
    a molecule, where no hydrogen is part of the transformation. See comments below on
    how this is achieved.

    Args:
        constant_smarts (str): constant smarts
        transformation_smarts (str): transformation smarts'
        molecule (Molecule): molecule to match smarts

    Raises:
        ValueError: Raised if a case occurrs that is not covered.

    Returns:
        tuple[list[int], ...] | tuple[list[list[int]], ...]: multiple or unique matches
        for constant and transformation smarts on  molecules.
    """
    # determine number of attachment points, needed to later identify
    # the correct transformation match, if more than one transformation
    # match is found.
    original_constant_smarts = constant_smarts
    num_attachment_points = 1
    for number in range(1, 4):
        if f"[*:{number}]" in constant_smarts:
            num_attachment_points = number
            constant_smarts = constant_smarts.replace(f"[*:{number}]", "")
            continue
        break
    constant_smarts = remove_stereo_at_beginning(smarts=constant_smarts)
    with contextlib.suppress(Exception):
        constant_matches: tuple[tuple[int]] = molecule.GetSubstructMatches(
            Chem.MolFromSmarts(  # type:ignore pylint: disable=E1101
                constant_smarts,
            ),
        )
    transformation_matches: tuple[tuple[int]] = molecule.GetSubstructMatches(
        Chem.MolFromSmarts(  # type:ignore pylint: disable=E1101
            transformation_smarts,
        ),
    )
    # start distinguishing different cases and try to get the correct match
    # for both the constant part as well as the transformation part.

    # this is a pretty bad case, since rdkit is not able to match a smarts
    # pattern that was created by an rdkit submodule. This will hopefully
    # be fixed on the rdkit side of thing in the near future.
    # an empty list for the constant match is returned, so things will be
    # taken care of on the circle level of things.

    if not constant_matches:
        return [[]], [list(match) for match in transformation_matches]

    # best possible case, both constant and transformation part are only
    # matched once. return these values

    if len(constant_matches) == len(transformation_matches) == 1:
        return list(constant_matches[0]), list(transformation_matches[0])

    # second best cases follow, by having either just one transformation
    # or constant match, the correct partner can be found by looking
    # for the pair that has an len(intersection of indices)==numattachmentpoints.

    if len(constant_matches) == 1 and len(transformation_matches) > 1:
        constant_match = list(constant_matches[0])
        transformation_match = get_correct_match(
            single_match=constant_match,
            multiple_matches=transformation_matches,
            num_attachment_points=num_attachment_points,
        )
        return constant_match, transformation_match

    if len(constant_matches) > 1 and len(transformation_matches) == 1:
        transformation_match = list(transformation_matches[0])
        constant_match = get_correct_match(
            single_match=transformation_match,
            multiple_matches=constant_matches,
            num_attachment_points=num_attachment_points,
        )
        return constant_match, transformation_match

    # now things get tricky. If both the constant as well as the
    # transformation pattern are matched more than once, there
    # are different cases that need to be distinguished, and
    # some that cannot be solved on this level of the classification
    # and will need reevaluation on the circle level of things.

    if len(constant_matches) > 1 and len(transformation_matches) > 1:
        # first case is going to be easy, this would be when
        # more than one chemically equivalent atom is attached
        # to the same atom and hence the constant part is
        # matched multiple times. In this case, you can take
        # any match since the important index is the one,
        # the r group is attached to which in every match,
        # should be the same.

        if all_matches_chemically_equivalent(
            matches=constant_matches,
            molecule=molecule,
        ):
            constant_match = list(constant_matches[0])
            transformation_match = get_correct_match(
                single_match=constant_match,
                multiple_matches=transformation_matches,
                num_attachment_points=num_attachment_points,
            )
            return constant_match, transformation_match

        # this second case is more worrysome, since here we are talking
        # cases that cannot be handled on the transformation level. here
        # we are talking cases were atoms are symetrically equivalent but
        # not bound to the same atom.

        if all_matches_symetrically_interchangeable(
            matches=constant_matches,
            molecule=molecule,
        ):
            return [list(m) for m in constant_matches], [
                list(m) for m in transformation_matches
            ]

    newmatch = molecule.GetSubstructMatches(
        Chem.MolFromSmarts(  # type:ignore pylint: disable=E1101
            original_constant_smarts,
        ),
    )
    if len(newmatch) == 1:
        matchset = set(newmatch[0])
        bettermatch = []
        for match in constant_matches:
            newset = set(match)
            if (
                len(newset.intersection(matchset))
                == len(newset) + num_attachment_points
            ):
                bettermatch = list(match)
                break
        transformation_match = get_correct_match(
            single_match=bettermatch,
            multiple_matches=transformation_matches,
            num_attachment_points=num_attachment_points,
        )
        return bettermatch, transformation_match
    return [list(m) for m in constant_matches], [
        list(m) for m in transformation_matches
    ]


def get_matches_with_hydrogen(
    constant_smarts: str,
    transformation_smarts: str,
    molecule: Molecule,
) -> tuple[list[int], ...] | tuple[list[list[int]], ...]:
    """Get match with hydrogen.

    This function tries to uniquely match transformation_smarts and constant_smarts in
    a molecule, where a hydrogen is part of the transformation. See comments below on
    how this is achieved.

    Args:
        constant_smarts (str): constant smarts
        transformation_smarts (str): transformation smarts'
        molecule (Molecule): molecule to match smarts

    Raises:
        ValueError: Raised if a case occurrs that is not covered.

    Returns:
        tuple[list[int], ...] | tuple[list[list[int]], ...]: multiple or unique matches
        for constant and transformation smarts on molecules.
    """

    def remove_stereochem(smarts: str) -> str:
        return smarts.replace("[*:1][C@H]", "[*:1]C").replace("[*:1][C@@H]", "[*:1]C")

    # same shenanigans but with hydrogens. since the transformation part
    # that is the hydrogens is matched all around the molecule, we leave
    # the [*:1] on the constant part first, so we can find the transformation
    # by looking for the match that has an len(intersection) == 2.
    mol_w_h = Chem.AddHs(molecule)  # type: ignore pylint: disable=E1101
    constant_smarts = remove_stereochem(constant_smarts)
    transformation_matches = mol_w_h.GetSubstructMatches(
        Chem.MolFromSmarts(transformation_smarts),  # type: ignore pylint: disable=E1101
    )
    constant_matches_w_h = mol_w_h.GetSubstructMatches(
        Chem.MolFromSmarts(constant_smarts),  # type: ignore pylint: disable=E1101
    )

    # what is actually returned in the end, since for the constant part
    # we only want the constant part later on.
    const_without_anchor = Chem.MolFromSmarts(  # type: ignore pylint: disable=E1101
        remove_stereo_at_beginning(constant_smarts.replace("[*:1]", "")),
    )
    if const_without_anchor is None or not const_without_anchor:
        return [[]], [list(m) for m in transformation_matches]
    constant_matches = molecule.GetSubstructMatches(
        const_without_anchor,
    )
    # Not able to solve this problem on this level of things,
    # since rdkit is not able to distinguish between chiral
    # hidrogens if two of them are attached to a carbon. This
    # is solved by looking at another compound on the circle level.

    if not constant_matches_w_h:
        return [[]], [list(m) for m in transformation_matches]

    # optimal case. find the correct hydrogen by looking for the match
    # that has a len(intersection) of 2

    if len(constant_matches_w_h) == 1:
        constant_match = list(constant_matches[0])
        transformation_match = get_correct_match(
            single_match=list(constant_matches_w_h[0]),
            multiple_matches=transformation_matches,
            num_attachment_points=2,
        )
        return constant_match, transformation_match

    # again gets tricky like it was the case without the hydrogens

    # if the different matches are all around the same CH3 for
    # example, we are only interested in the C.

    if (
        all_matches_chemically_equivalent(
            matches=constant_matches_w_h,
            molecule=mol_w_h,
        )
        and len(constant_matches) == 1
    ):
        constant_match = list(constant_matches[0])
        transformation_match = get_correct_match(
            single_match=list(constant_matches_w_h[0]),
            multiple_matches=transformation_matches,
            num_attachment_points=2,
        )
        return constant_match, transformation_match

    # again the case, where we have symetrically equivalent atoms,
    # we need to solve the problem on the next higher level, which
    # is going to be the circle level.

    if all_matches_symetrically_interchangeable(
        matches=constant_matches_w_h,
        molecule=mol_w_h,
    ):
        return [list(m) for m in constant_matches_w_h], [
            list(m) for m in transformation_matches
        ]

    return [[]], [list(m) for m in transformation_matches]


def get_constant_and_transformation_indices(
    transformation_smarts: str,
    constant_smarts: str,
    molecule: Molecule,
) -> tuple[list, ...]:
    """Identify constant and trasnformation indices.

    Tries to identify which atoms in a molecule are part of the constant part
    and the transformation part of a transformation.

    Args:
        transformation_smarts (str): SMARTS of fragment that is exchanged
        constant_smarts (str): SMARTS of fragment that is constant
        molecule (Molecule): molecule to match

    Returns:
        tuple[list, ...]: tuple of constant atom indices and transformation atom
        indices. If match is ambiguos, lists of lists or an empty list of list is
        returned.
    """
    if transformation_smarts == "[*:1][H]":
        return get_matches_with_hydrogen(
            constant_smarts=constant_smarts,
            transformation_smarts=transformation_smarts,
            molecule=molecule,
        )
    return get_matches_without_hydrogen(
        constant_smarts=constant_smarts,
        transformation_smarts=transformation_smarts,
        molecule=molecule,
    )


def get_r_group_anchor(
    constant_indices: list,
    transformation_indices: list,
) -> list[int] | tuple[list, ...]:
    """Get atom indices of r group anchors.

    Returns the atom indices of the transformation molecules
    to which the r groups are connected to.

    Returns:
        list[int] | tuple[list, ...]: r-group connection indices.
    """
    if not all(is_match_unique(m) for m in (constant_indices, transformation_indices)):
        return constant_indices, transformation_indices
    intersection = set(constant_indices).intersection(transformation_indices)
    if not intersection:
        raise ValueError(
            f"not intersection\
                \nconstant_indices {constant_indices}\
                \transformation_indices {transformation_indices}",
        )
    return list(intersection)


# -------------------------------------------------------------------------
#
#                       CIRCLE LEVEL OF THINGS
#
# =========================================================================


def try_get_num_atoms_between_rgroups(
    transformation_1: Transformation,
    transformation_2: Transformation,
) -> float:
    """Try get number of atoms between two r groups.

    Tries to determine the distance between two R-Groups that are
    changed respectively in transformation_1 and transformation_2.

    Args:
        transformation_1 (Transformation): First Transformation
        transformation_2 (Transformation): Second Transformation

    Raises:
        NotImplementedError: Thrown if a case i could not think of
            yet is thrown.

    Returns:
        int: -1 if transformation 1 is not unambiguosly defined
             -2 if transformation 2 is not unambiguosly defined
             -3 if both transformations are not unambiguosly defined
            >=0 actual number of bonds between rgroups
    """

    def get_atoms_between(mol: Molecule, index_1: int, index_2: int) -> int:
        return (
            1
            if index_1 == index_2
            else len(Chem.GetShortestPath(mol, index_1, index_2))
        )

    anchor_1, anchor_2 = (
        transformation_1.compound_1_anchor,
        transformation_2.compound_1_anchor,
    )
    molecule = transformation_1.compound_1.rdkit_molecule
    anchor_1_is_unique = isinstance(anchor_1, list)
    anchor_2_is_unique = isinstance(anchor_2, list)
    if not anchor_1_is_unique and anchor_2_is_unique:
        return -1
    if anchor_1_is_unique and not anchor_2_is_unique:
        return -2
    if not all((anchor_1_is_unique, anchor_2_is_unique)):
        return -3

    if len(anchor_1) == len(anchor_2) == 1:
        min_distance = get_atoms_between(
            mol=molecule,
            index_1=anchor_1[0],
            index_2=anchor_2[0],
        )
    else:
        min_distance = float("inf")
        for i in anchor_1:
            for j in anchor_2:
                distance = get_atoms_between(
                    mol=molecule,
                    index_1=i,
                    index_2=j,
                )
                if distance < min_distance:
                    min_distance = distance
    if (
        min_distance == 2
        and len(
            set(transformation_1.compound_1_transformation_indices).intersection(
                set(transformation_2.compound_1_transformation_indices),
            ),
        )
        == 2
    ):
        return 0
    return min_distance


def get_num_atoms_between_rgroups(**kwargs: Transformation) -> int:
    """Calculate number of atoms between r groups in a double transformation cycle.

    following keyword arguments are needed:
        'transformation_1',
        'transformation_2',
        'transformation_3',
        'transformation_4',

    Raises:
        NotImplementedError: thrown if a case occurs that is not implemented yet.

    Returns:
        int: number of atoms between r groups.
    """
    distance = -1
    for pot_transf_1, pot_transf_2 in get_combinations(**kwargs):
        distance = try_get_num_atoms_between_rgroups(
            transformation_1=pot_transf_1,
            transformation_2=pot_transf_2,
        )
        if distance >= 0:
            break
    return int(distance)



"""Implement circle classification.

Implements all the classification functions called in classification_classes:Circle
that are not responsible for the ortho or the r group distance classification.
"""

from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING

import numpy as np

from nonadditivity.classification.utils import is_unique

if TYPE_CHECKING:
    from nonadditivity.classification import (
        Compound,
        Transformation,
    )
    from nonadditivity.utils.types import Molecule


AMINE_DICT = {-1: "Not identifiable", 0: "False", 1: "True"}
STEREO_DICT = {0: "None", 1: "unassigned", 2: "inversion"}


def _any_compound_has_property(
    property_key: Compound.Properties,
    **kwargs: Transformation,
) -> bool:
    """Return true if any of the compounds have true for <property_key>.

    kwargs 'transformation_1', 'transformation_2', 'transformation_3' and
    'transformation_4' needed for this to work.

    Args:
        property_key (Compound.Properties): property to look at.
        kwargs: keyword arguments.

    Returns:
        bool: True if any any of the compounds have True at <property_key>
    """
    compound_1 = kwargs["transformation_1"].compound_1
    compound_2 = kwargs["transformation_1"].compound_2
    compound_3 = kwargs["transformation_4"].compound_2
    compound_4 = kwargs["transformation_4"].compound_1
    return any(
        c.get_property(property_key)
        for c in (compound_1, compound_2, compound_3, compound_4)
    )


def _get_max_in_circle_transformations(
    property_key: Transformation.Properties,
    iterators: tuple[int, ...],
    **kwargs: Transformation,
) -> int:
    """Get for maximum value of property for transformations in iterators.

    Returns maximum value for a <property_key> for any of the transformations
    given in <iterators> for a given circle.

    Args:
        property_key (Transformation.Properties): property to get max value for.
        iterators (tuple[int]): which transformations to look at.
        kwargs: kwargs to use in this.

    Returns:
        int: max value found in <iterators> transformations for <property_key>
    """
    max_val = [
        t.get_property(property_key)
        for t in [kwargs[f"transformation_{i}"] for i in iterators]
    ]
    if isinstance(max_val[0], Iterable):
        max_val = list(np.concatenate(max_val))
    return max(max_val)


def get_compound_stereocenter_classification(**kwargs: Transformation) -> str:
    """Get stereo classification for compounds in circle.

    use kwargs transformation_1, 2, 3 and 4 for this to work.

    Returns:
        str: stereo classification for compounds
    """
    prop = kwargs["transformation_1"].compound_1.Properties
    if not _any_compound_has_property(prop.NUM_STERO_CENTERS, **kwargs):
        return "None"
    return (
        "Unassigned"
        if _any_compound_has_property(prop.HAS_UNASSIGNED_STEREOCENTERS, **kwargs)
        else "Assigned"
    )


def circle_has_stereoinversion_in_transformation(**kwargs: Transformation) -> str:
    """Check whether a transformation in the circle is a stereoinversion.

    Checks whether any transformation in a circle is only a stereoinversion,
    resp. a change from assigned to unassigned stereochemistry.

    use kwargs transformation_1 and transformation_2 for this to work.

    Returns:
        str: 'None' == transformations are not stereochemistry related
            'unassigned' == at least one transformation transforms an assigned
            stereo isomer to an unassigned stereoisomer.
            'inversion' == at least one transformation is a stereoinversion
    """
    prop = kwargs["transformation_1"].Properties.TRANSFORMATION_STEREO
    return STEREO_DICT[
        _get_max_in_circle_transformations(
            property_key=prop,
            iterators=(1, 2),
            **kwargs,
        )
    ]


def get_tertiary_amide_formed(**kwargs: Transformation) -> str:
    """Check whether any transformation in a circle creates a tertiary amide.

    use kwargs transformation_1, 2, 3 and 4 for this to work.

    Returns:
        str: 'True' == tertiary amide formed in any of the transformations
            'False' == no tertiary amide formed in any of the transformations
            'Not Identifiable' == not able to perform check.
    """
    prop = kwargs["transformation_1"].Properties.TERTIARY_AMIDE_FORMED
    return AMINE_DICT[
        _get_max_in_circle_transformations(
            property_key=prop,
            iterators=(1, 2, 3, 4),
            **kwargs,
        )
    ]


def get_max_num_heavy_atom_in_transformation(**kwargs: Transformation) -> int:
    """Return biggest transformation of a circle.

    use kwargs transformation_1 and transformation_2 for this to work.

    Returns:
        int: highest number of atoms found in any r group.
    """
    prop = kwargs["transformation_1"].Properties.NUM_HEAVY_ATOMS_IN_RGROUPS
    return _get_max_in_circle_transformations(
        property_key=prop,
        iterators=(1, 2),
        **kwargs,
    )


def get_max_num_mmpdb_cuts(**kwargs: Transformation) -> int:
    """Returns highest number of mmpdb cuts in a circle.

    use kwargs transformation_1 and transformation_2 for this to work.

    Returns:
        int: highest num mmpdb cuts in a circle.
    """
    prop = kwargs["transformation_1"].Properties.NUM_MMPDB_CUTS
    return _get_max_in_circle_transformations(
        property_key=prop,
        iterators=(1, 2),
        **kwargs,
    )


def get_min_transformation_tanimoto(**kwargs: Transformation) -> float:
    """Get smaller tanimoto of fingerprints for either t1 and t4 or t2 and t3.

    use kwargs transformation_1, 2, 3 and for this method to work.

    Returns:
        float: mean of tanimoto of fingerprints of transformation1 and 4
    """
    t1 = kwargs["transformation_1"]
    t4 = kwargs["transformation_4"]
    t3 = kwargs["transformation_3"]
    t2 = kwargs["transformation_2"]
    return min(
        np.mean(
            [
                t2.get_property(t2.Properties.MFP2_SIMILARITY),
                t3.get_property(t3.Properties.MFP2_SIMILARITY),
            ],
        ),
        np.mean(
            [
                t1.get_property(t1.Properties.MFP2_SIMILARITY),
                t4.get_property(t4.Properties.MFP2_SIMILARITY),
            ],
        ),
    )


def __calc_max_delta(
    val_1: float,
    val_2: float,
    val_3: float,
    val_4: float,
) -> float:
    """Get biggest absolute difference in (val1-val3) and (val2 - val4).

    Args:
        val_1 (float): val1
        val_2 (float): val2
        val_3 (float): val3
        val_4 (float): val4

    Returns:
        float: max(abs(v1-v3),abs(v2-v4))
    """
    return max(
        abs(
            val_1 - val_3,
        ),
        abs(
            val_2 - val_4,
        ),
    )


def _get_max_diagonal_delta_from_compounds(
    property_key: Compound.Properties,
    **kwargs: Transformation,
) -> float:
    """Find the biggest delta in property in a double transformation cycle.

    use kwargs transformation_1, 2, 3 and 4 for this to work.

    Args:
        property_key (Transformation.Properties): name of the transformation property
        kwargs: kwargs needed for this function.

    Returns:
        float: max(abs(diff_1,2)) for diff_1,2 being the two diagonal differencees.
    """
    val_1 = kwargs["transformation_1"].compound_1.get_property(property_key)
    val_2 = kwargs["transformation_1"].compound_2.get_property(property_key)
    val_3 = kwargs["transformation_4"].compound_2.get_property(property_key)
    val_4 = kwargs["transformation_4"].compound_1.get_property(property_key)

    return __calc_max_delta(val_1, val_2, val_3, val_4)


def get_hbond_donor_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in h bond donors for double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in h bond donors
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.NUM_HBD
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_hbond_acceptor_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in h bond acceptors for double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in h bond acceptors
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.NUM_HBA
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_formal_charge_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in formal charge for double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in formal charge
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.CHARGE
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_tpsa_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in tpsa for double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in tpsa
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.TPSA
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_num_rot_bonds_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in num of rotatable bonds for double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in number of rotatable bonds
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.NUM_ROT_BONDS
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_sp3_carbon_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in num of spc 3 carbons for double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in spc 3 carbon fraction
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.NUM_SP3_CARBONS
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_log_p_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in log p for double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in log p
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.LOGP
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_num_heavy_atoms_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in heavy atoms double transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in heavy atoms
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.NUM_HEAVY_ATOMS
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_chi0_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in chi1 in the 4 r groups in  transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in chi1
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.CHI0
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_chi1_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in chi1 in the 4 r groups in  transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in chi1
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.CHI1
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_chi2_diff(**kwargs: Transformation) -> float:
    """Get biggest difference in chi1 in the 4 r groups in  transformation cycle.

    use keyword arguments 'transformation_1', 'transformation_2'

    Returns:
        float: biggest circle diff in chi1
    """
    property_key = kwargs["transformation_1"].compound_1.Properties.CHI2
    return _get_max_diagonal_delta_from_compounds(property_key, **kwargs)


def get_combinations(
    **kwargs: Transformation,
) -> list[tuple[Transformation, Transformation]]:
    """Get all for combinations of transformations to describe a compound.

    use keyword arguments transformation_1 ,2,3,4 for this to work.

    Returns:
        list[tuple[Transformation, Transformation]]: transformations
    """
    t1 = kwargs["transformation_1"]
    t2 = kwargs["transformation_2"]
    t3 = kwargs["transformation_3"]
    t4 = kwargs["transformation_4"]
    return [
        (t1, t2),
        (t1.get_reverse(), t3),
        (t2.get_reverse(), t4),
        (t3.get_reverse(), t4.get_reverse()),
    ]


def _substituents_in_same_ring_system(
    molecule: Molecule,
    index_1: int,
    index_2: int,
    aromatic_indices: set[int],
    fused_rings: list[set[int]],
) -> bool:
    ringinfo = molecule.GetRingInfo()
    if ringinfo.AreAtomsInSameRing(index_1, index_2) and all(
        i in aromatic_indices for i in (index_1, index_2)
    ):
        return True
    return any(all(v in ring for v in (index_1, index_2)) for ring in fused_rings)


def substituents_in_same_ring_system(**kwargs: Transformation) -> bool:
    """Check whether the two r groups in a cricle are connected to the same ring.

    Use keyword arguments transformation_1, 2, 3 and 4 for this to work.

    Returns:
        bool: True if both r groups are attached to the same ring.
    """
    for pot_transf_1, pot_transf_2 in get_combinations(**kwargs):
        anchor_1 = pot_transf_1.compound_1_anchor
        anchor_2 = pot_transf_2.compound_1_anchor
        compound = pot_transf_1.compound_1
        if not all(is_unique(a) for a in (anchor_1, anchor_2)):
            continue
        if not all(len(a) == 1 for a in (anchor_1, anchor_2)):
            continue
        return _substituents_in_same_ring_system(
            molecule=compound.rdkit_molecule,
            index_1=anchor_1[0],
            index_2=anchor_2[0],
            aromatic_indices=compound.get_property(
                compound.Properties.AROMATIC_INDICES,
            ),
            fused_rings=compound.get_property(compound.Properties.FUSED_RING_INDICES),
        )
    return False



"""Math utils for the Nonadditivity Analysis package."""

from collections.abc import Sequence
from typing import Any

import numpy as np
from pandas import DataFrame
from scipy import stats


def is_number(value: Any) -> bool:
    """Check whether value is convertible to a float.

    Args:
        value (Any): value to check.

    Returns:
        bool: true if value is a number and is or is convertible to float.
    """
    try:
        float(value)
        return True
    except (ValueError, TypeError):
        return False


def mad_std(values: Sequence[float]) -> float:
    """Calculate median absolute deviation.

    Estimate the standard deviation based on the median absolute deviation
    (MAD) In Gaussian distributions, SD = 1.4826 * MAD

    This is a fast and simple robust SD estimator. However, it may have
    problems for small datasets where the average is not well estimated.

    Args:
        values (Sequence[float]): values to compute the MADSTD

    Returns:
        float: MADSTD for values

    Raises:
        ValueError: if either wrong type, e.g. NoneType or wrong values e.g.
        list[str] are given.
    """
    try:
        return float(1.4826 * np.median(np.abs(values - np.median(values))))
    except (ValueError, TypeError) as exc:
        raise ValueError(
            "Values has to be of type np.ndarray[float] or list[float]",
        ) from exc


def sn_medmed_std(values: Sequence[float]) -> float:
    """Calculate std off median pairwise difference.

    Estimate the standard deviation based on the median of median pairwise
    differences as defined by Rousseeuw and Croux [1]:

    Sn = 1.1926 * med_i( med_j(abs(xi - xj)))

    [1] Rousseeuw, Peter J.; Croux, Christophe (December 1993),
    "Alternatives to the Median Absolute Deviation",
    Journal of the American Statistical Association, American Statistical
    Association, 88 (424): 1273-1283, doi:10.2307/2291267, JSTOR 2291267

    Returns 0 if len(values) == 1

    Args:
        values (Sequence[float]): values to compute the sn_medmed_std

    Returns:
        float:  sn_medmed_std for vlaues

    Raises:
        ValueError: if wrong values or types given.
    """
    try:
        iter(values)
    except TypeError as exc:
        if not is_number(values):
            raise ValueError(
                "Values has to be of type np.ndarray[float] or list[float]",
            ) from exc
        return 0
    try:
        pairwise_medians = np.empty(len(values))
        for idx, i_value in enumerate(values):
            # Do I need to remove the case where i==j ? Tests indicate yes
            pairwise_medians[idx] = np.median(
                np.abs(i_value - np.delete(values, [idx])),
            )
        return float(1.1926 * np.median(pairwise_medians))
    except (ValueError, TypeError) as exc:
        raise ValueError(
            "Values has to be of type np.ndarray[float] or list[float]",
        ) from exc


def _calculate_series_theo_quantile(
    series_ids: list[str],
    nonadditivities: list[float],
) -> list[list[float]]:
    """Calculate theoquantiles using scipy.stats.boxplot.

    Args:
        series_ids (list[str]): list of the series name per circle
        nonadditivities (list[float]): nonadditivitiy per circle.

    Returns:
        list[list[float]]: list of the theoquantiles per circle
    """
    theo_quantiles = []
    for series in set(series_ids):
        sorted_nonadditivitiies = sorted(
            [
                (nonadditivity, x[0])
                for nonadditivity, x in zip(
                    nonadditivities,
                    enumerate(series_ids),
                )
                if x[1] == series
            ],
        )
        additivity_indices = [x[1] for x in sorted_nonadditivitiies]
        if len(additivity_indices) > 2:
            quantiles = list(stats.probplot(additivity_indices)[0][0])
        elif len(additivity_indices) == 2:
            quantiles = [-0.54495214, 0.54495214]
        else:
            quantiles = [0, 0]
        theo_quantiles += list(zip(additivity_indices, quantiles))
    return [quantile[1] for quantile in sorted(theo_quantiles)]


def _calculate_theo_quantiles(
    series_columns: str | None,
    series_ids: list[str],
    nonadditivities: list[float],
) -> list[list[float]]:
    """Calculate theoquantiles using scipy.stats.probplot.

    Args:
        series_columns (str | None): name of the series column.
        series_ids (list[str]): list of the series name per circle
        nonadditivities (list[float]): nonadditivitiy per circle.

    Returns:
        list[list[float]]: list of the theoquantiles per circle
    """
    if series_columns:
        return _calculate_series_theo_quantile(series_ids, nonadditivities)
    sorted_nonadditivities = sorted(
        zip(nonadditivities, range(len(nonadditivities))),
    )
    theo_quantiles = stats.probplot(sorted(nonadditivities))[0][0]
    theo_quantiles = [
        j[1]
        for j in sorted(
            zip(
                [i[1] for i in sorted_nonadditivities],
                list(theo_quantiles),
            ),
        )
    ]
    return theo_quantiles


def _calculate_mean_and_std_per_compound(
    per_compound_dataframe: DataFrame,
    property_column: str,
    series_column: str | None = None,
) -> list[list[float | None]] | list[list[list[float | None]]]:
    """Calculate mean nonadditivty and std per compound.

    Takes all nonadditivity values per compound and calculates mean,
    std and number of occurence.

    Args:
        per_compound_dataframe (DataFrame): per compound dataframe.
        property_column (str): name of the property column
        series_column (str | None, optional): name of the series column.
        Defaults to None.

    Returns:
        list[float | None] | list[list[float | None]]: list
        of lists containing mean, std, n_occurence for every row in the per
        compound dataframe.
    """

    def calculate_values(values: list[float]) -> list[float | None]:
        """Calculate mean std and num values for a list of values.

        Returns [None, None, 0] if the list is empty.

        Args:
            values (list[float | None]): list of mean, std and num values.

        Returns:
            list[Union[float, None]]: [mean, std, num_values] if len(values)
            > 0 else [None, None, 0]
        """
        num_values = len(values)
        if num_values > 0:
            mean = np.mean(values)
            std = np.std(values)
            return [float(mean), float(std), float(num_values)]
        return [None, None, 0]

    column = f"{property_column}_Nonadditivities"
    if series_column:
        pure_values, mixed_values = [], []
        for nonadditivities in per_compound_dataframe[column].to_numpy():
            pure_values.append(
                calculate_values(
                    values=[
                        value[0] for value in nonadditivities if value[1] == "pure"
                    ],
                ),
            )
            mixed_values.append(
                calculate_values(
                    values=[
                        value[0] for value in nonadditivities if value[1] == "mixed"
                    ],
                ),
            )
        return [pure_values, mixed_values]
    return [
        calculate_values(
            values=[value[0] for value in nonadditivities],
        )
        for nonadditivities in per_compound_dataframe[column].to_numpy()
    ]


def _calculate_nonadditivity(
    values: list[float],
) -> float:
    r"""Calculate the actual nonadditivity value as defined in the paper.

    C1-C2
    |	|
    C4-C3

    Na = p(C3) + p(C1) - p(C2) - p(C4)

    Args:
        values (list[float]): Values for the 4 compounds

    Returns:
        float: Calculated Nonaddictivity
    """
    return values[2] + values[0] - values[1] - values[3]



"""Canonicalizing nonadditivity output.

Helper Module that canonicalizes output files in a way,
that there are no more transformations in the output,
that are the reverse of another transformation.
"""

import contextlib

import pandas as pd

from nonadditivity.classification import Circle
from nonadditivity.utils.math import _calculate_nonadditivity
from nonadditivity.workflow.nonadditivity_core import _create_circle_ids


def _get_reverse(transformation: str) -> str:
    """Get reverse transformation: 'A>>B' is transformed to 'B>>A'.

    Args:
        transformation (str): transformation smarts

    Returns:
        str: reversed transformation smarts

    Raises:
        ValueError: Raised if '>>' not in transformation.
    """
    if ">>" not in transformation:
        raise ValueError("transformation has to contain '>>'")
    parts = transformation.split(">>")
    return parts[1] + ">>" + parts[0]


def _get_correct_transformation_smarts(
    t1: str,
    t2: str,
    t1_reversed: bool,
    t2_reversed: bool,
) -> tuple[str, ...]:
    """Get the correct version of t1 and t2.

    reverses them if t1_changed = True, resp. for t2.

    Args:
        t1 (str): transformation 1
        t2 (str): transformation 2
        t1_reversed (bool): whether transformation 1 should be reversed.
        t2_reversed (bool): whether transformation 2 should be reversed.

    Returns:
        tuple[str, ...]: correct transformations.
    """
    return (
        _get_reverse(t1) if t1_reversed else t1,
        _get_reverse(t2) if t2_reversed else t2,
    )


def _get_correct_order(
    t1_reversed: bool,
    t2_reversed: bool,
) -> tuple[int, ...]:
    """Get correct index order for canonical transformation.

    Transforms the indices (1,2,3,4) according to wheter any or all of the
    transformations are reversed.

    Args:
        t1_reversed (bool): transformation1 is reversed
        t2_reversed (bool): transformation2 is reversed

    Returns:
        tuple[int, ...]: new order of the circle
    """
    if all([t1_reversed, t2_reversed]):
        return (3, 4, 1, 2)
    if t1_reversed:
        return (2, 1, 4, 3)
    if t2_reversed:
        return (4, 3, 2, 1)
    return (1, 2, 3, 4)


def _get_props() -> list[str]:
    """Return circle porperty names.

    Returns all Properties in nonadditivity.classification_classes.Circle
    that are not ortho_classification or num_bonds_between_atoms.

    Returns:
        list[str]: list of descripytion keys
    """
    return (
        ["num_atoms_between_r_groups", "ortho_classification"]
        + [
            val
            for val in Circle.classification_keys.values()
            if "ortho" not in val and "between" not in val
        ]
        + ["classification"]
    )


def _add_to_dict(
    row: pd.Series,
    index: int,
    df_dict: dict[str, list | pd.Series],
    t1_reversed: bool,
    t2_reversed: bool,
) -> None:
    """If necessary reverse transformations and append the new values to df_dict.

    Args:
        row (pd.Series): row to be transformed
        index (int): index of this row in dataframe
        df_dict (dict[str, list]): dict to add values to
        t1_reversed (bool): whether t1 has to be reversed
        t2_reversed (bool): whether t2 has to be reversed
    """
    order = (1, 2, 3, 4)
    transformation1, transformation2 = (
        row["Transformation1"],
        row["Transformation2"],
    )
    if any([t1_reversed, t2_reversed]):
        new_order = _get_correct_order(t1_reversed, t2_reversed)
        transformation1, transformation2 = _get_correct_transformation_smarts(
            transformation1,
            transformation2,
            t1_reversed,
            t2_reversed,
        )
    else:
        new_order = order
    for entry, new in zip(order, new_order):
        for name in ("Compound", "SMILES", "Prop_Cpd"):
            df_dict[f"{name}{entry}"].append(row[f"{name}{new}"])

    df_dict["Property"].append(row["Property"])
    with contextlib.suppress(KeyError):
        df_dict["Series"].append(row["Series"])

    df_dict["Circle_ID"].append(
        _create_circle_ids(
            [[df_dict[f"Compound{i}"][index] for i in range(1, 5)]],
            row["Property"],
        )[0],
    )
    old_na = row["Nonadditivity"]
    new_na = _calculate_nonadditivity(
        values=[row[f"Prop_Cpd{i}"] for i in new_order],
    )
    df_dict["Nonadditivity"].append(new_na)
    theo_sign = 1 if old_na == new_na else -1
    df_dict["Theo_Quantile"].append(row["Theo_Quantile"] * theo_sign)
    df_dict["Transformation1"].append(transformation1)
    df_dict["Transformation2"].append(transformation2)

    for prop in _get_props():
        try:
            df_dict[prop].append(row[prop])
        except KeyError:
            continue


def _create_df_dict(
    na_dataframe: pd.DataFrame,
) -> dict[str, list | pd.Series]:
    """Create dict with unique transformations.

    Creates a dictionary containing all na_dataframes values but with only unique
    transformations.

    Args:
        na_dataframe (pd.DataFrame): dataframe to be canonicalized

    Returns:
        dict[str, list | pd.Series]: dict[columns, canonicalized values]
    """
    df_dict: dict[str, list | pd.Series] = {
        str(key): [] for key in na_dataframe.columns.to_numpy()
    }
    seen = set()
    for index, (_, row) in enumerate(na_dataframe.iterrows()):
        t1_changed, t2_changed = False, False
        t1, t2 = row["Transformation1"], row["Transformation2"]
        if t1 not in seen:
            if _get_reverse(t1) in seen:
                t1_changed = True
            else:
                seen.add(t1)
        if t2 not in seen:
            if _get_reverse(t2) in seen:
                t2_changed = True
            else:
                seen.add(t2)
        _add_to_dict(row, index, df_dict, t1_changed, t2_changed)
    return df_dict


def canonicalize_na_dataframe(
    na_dataframe: pd.DataFrame,
) -> pd.DataFrame:
    """Canonicalize nonadditivity dataframe.

    Creates a dataframe that contains only canonicalized transformations,
    i.e. only one of the following is present: 'Cl>>F', 'F>>Cl'

    Args:
        na_dataframe (pd.DataFrame): dataframe to be transformed

    Returns:
        pd.DataFrame: canonicalized dataframe.
    """
    canonical_df = pd.DataFrame()
    try:
        na_dataframe_fixed = na_dataframe.copy().drop(
            columns=[
                "has_transformation_at_ortho",
                "has_ortho_substituent_introduced",
            ],
        )
    except KeyError:
        na_dataframe_fixed = na_dataframe
    df_dict = _create_df_dict(na_dataframe_fixed)
    for col in na_dataframe_fixed.columns.to_numpy():
        if col not in df_dict:
            continue
        canonical_df[col] = df_dict[col]
    return canonical_df


def get_transformation_df(canonical_df: pd.DataFrame) -> pd.DataFrame:
    """Create dataframe with canonical trasnformations.

    Creates a dataframe containing all the transformations and the respective
    activity values.

    Args:
        canonical_df (pd.DataFrame): input dataframe

    Returns:
        pd.DataFrame: df containing 'Property', 'Transformation', 'Val1', 'Val2'.
    """
    transformations_dataframe = pd.DataFrame()
    transformations = []
    val1, val2 = [], []
    targets = []

    for tr1, tr2, prop1, prop2, prop3, prop4, target in zip(
        canonical_df.Transformation1.to_numpy(),
        canonical_df.Transformation2.to_numpy(),
        canonical_df.Prop_Cpd1.to_numpy(),
        canonical_df.Prop_Cpd2.to_numpy(),
        canonical_df.Prop_Cpd3.to_numpy(),
        canonical_df.Prop_Cpd4.to_numpy(),
        canonical_df.Property.to_numpy(),
    ):
        for _ in range(4):
            targets.append(target)
        transformations.append(tr1)
        val1.append(prop1)
        val2.append(prop2)
        transformations.append(tr1)
        val1.append(prop4)
        val2.append(prop3)
        transformations.append(tr2)
        val1.append(prop1)
        val2.append(prop4)
        transformations.append(tr2)
        val1.append(prop2)
        val2.append(prop3)
    for props in _get_props():
        try:
            transformations_dataframe[props] = [
                canonical_df.loc[index, props]
                for index in canonical_df.index.to_numpy()
                for _ in range(4)
            ]
        except KeyError:
            continue
    transformations_dataframe["Property"] = targets
    transformations_dataframe["Transformation"] = transformations
    transformations_dataframe["Val1"] = val1
    transformations_dataframe["Val2"] = val2
    with contextlib.suppress(KeyError):
        transformations_dataframe["Series"] = [
            series for series in canonical_df["Series"].to_numpy() for _ in range(4)
        ]
    return transformations_dataframe



"""Implement Compound class for classification.

Holds Class for Describing Properties on a compound Level.
This is used for the NonadditivityAnalysis Package.
"""

from collections.abc import Callable
from enum import Enum, auto
from typing import ClassVar

from rdkit import Chem

from nonadditivity.classification.compound_classification import (
    get_aromatic_indices,
    get_charge,
    get_chi0,
    get_chi1,
    get_chi2,
    get_fused_ring_indices,
    get_log_p,
    get_morgan_fp,
    get_num_h_bond_acceptors,
    get_num_h_bond_donors,
    get_num_heavy_atoms,
    get_num_rotatable_bonds,
    get_num_sp3_carbons,
    get_num_stereo_centers,
    get_polar_surface_area,
    has_unassigned_stereocenters,
)
from nonadditivity.classification.ortho_classification import (
    get_num_ortho_configurations,
    get_ortho_indices,
)
from nonadditivity.utils.types import Molecule


class Compound:
    """Compound class holds an compound_Id, smiles and a rdkit molecule.

    It also calculates the properties enumerated in Properties
    """

    class Properties(Enum):
        """Enumerates type of compound classifications."""

        NUM_HEAVY_ATOMS = auto()
        NUM_ORTHO_CONFIGURATIONS = auto()
        HAS_UNASSIGNED_STEREOCENTERS = auto()
        NUM_STERO_CENTERS = auto()
        ORTHO_INDICES = auto()
        AROMATIC_INDICES = auto()
        FUSED_RING_INDICES = auto()
        CHARGE = auto()
        NUM_HBA = auto()
        NUM_HBD = auto()
        TPSA = auto()
        LOGP = auto()
        CHI0 = auto()
        CHI1 = auto()
        CHI2 = auto()
        NUM_ROT_BONDS = auto()
        NUM_SP3_CARBONS = auto()
        MORGANFP = auto()

    classification_function: ClassVar[dict[Properties, Callable]] = {
        Properties.NUM_HEAVY_ATOMS: get_num_heavy_atoms,
        Properties.NUM_STERO_CENTERS: get_num_stereo_centers,
        Properties.NUM_ORTHO_CONFIGURATIONS: get_num_ortho_configurations,
        Properties.ORTHO_INDICES: get_ortho_indices,
        Properties.AROMATIC_INDICES: get_aromatic_indices,
        Properties.FUSED_RING_INDICES: get_fused_ring_indices,
        Properties.HAS_UNASSIGNED_STEREOCENTERS: has_unassigned_stereocenters,
        Properties.CHARGE: get_charge,
        Properties.NUM_HBA: get_num_h_bond_acceptors,
        Properties.NUM_HBD: get_num_h_bond_donors,
        Properties.TPSA: get_polar_surface_area,
        Properties.LOGP: get_log_p,
        Properties.CHI0: get_chi0,
        Properties.CHI1: get_chi1,
        Properties.CHI2: get_chi2,
        Properties.NUM_ROT_BONDS: get_num_rotatable_bonds,
        Properties.NUM_SP3_CARBONS: get_num_sp3_carbons,
        Properties.MORGANFP: get_morgan_fp,
    }

    classification_keys: ClassVar[dict[Properties, str]] = {
        Properties.NUM_HEAVY_ATOMS: "num_heavy_atoms",
        Properties.NUM_STERO_CENTERS: "num_stereo_centers",
        Properties.NUM_ORTHO_CONFIGURATIONS: "num_ortho_configurations",
        Properties.ORTHO_INDICES: "ortho_indices",
        Properties.AROMATIC_INDICES: "aromatic_indices",
        Properties.FUSED_RING_INDICES: "fused_ring_indices",
        Properties.HAS_UNASSIGNED_STEREOCENTERS: "has_unassigned_stereocenters",
        Properties.CHARGE: "charge",
        Properties.NUM_HBA: "num_hba",
        Properties.NUM_HBD: "num_hdb",
        Properties.TPSA: "tpsa",
        Properties.LOGP: "log_p",
        Properties.CHI0: "chi0",
        Properties.CHI1: "chi1",
        Properties.CHI2: "chi2",
        Properties.NUM_ROT_BONDS: "num_rotatable_bonds",
        Properties.NUM_SP3_CARBONS: "num_sp3_carbons",
        Properties.MORGANFP: "morgan_fp",
    }

    def __init__(
        self,
        molecule: Molecule,
        compound_id: str,
        smiles: str,
    ) -> None:
        """Instanciate compound object holding an id and an rdkit molecule and smiles.

        Args:
            molecule (Molecule): rdkit molecule
            compound_id (str): compound id
            smiles (str): compound smiles
        """
        self.rdkit_molecule = Chem.MolFromSmiles(  # type: ignore pylint: disable=E1101
            Chem.MolToSmiles(molecule),  # type: ignore pylint: disable=E1101
        )
        self.compound_id = compound_id
        self.smiles = smiles
        self._classification: dict[
            Compound.Properties,
            int | list[tuple[set[int], ...]],
        ] = {}

    def get_property(
        self,
        key: Properties,
    ) -> int | list[tuple[set[int], ...]] | set[int] | list[set[int]]:
        """Get a Property of a compound, all in Compound.Properties are allowed.

        Args:
            key (CompoundProperties): property name

        Returns:
            int: compound porperty
        """
        try:
            return self._classification[key]
        except KeyError:
            self._classification[key] = Compound.classification_function[key](
                molecule=self.rdkit_molecule,
            )
            return self.get_property(key=key)

    def classify(self) -> None:
        """Calculate all properties enumerated in.Compound.Properties."""
        for prop in Compound.Properties:
            if prop not in self._classification:
                self.get_property(prop)

    def get_classification_dict(
        self,
        force_classification: bool = True,
    ) -> dict[Properties, int | list[tuple[set[int], ...]]]:
        """Get dictionary with the properties already calculated for the compound.

        Args:
            force_classification (bool, optional): Forces calculation of all properties.
            Defaults to True.

        Returns:
            dict[str, int]: calculated properties
        """
        if force_classification:
            self.classify()
        return self._classification

    def get_formatted_classification_dict(
        self,
        force_classification: bool = True,
    ) -> dict[str, int | list[tuple[set[int], ...]]]:
        """Return dict with key for datframe and calculated properties.

        Return a dictionary with the properties already calculated for the compound
        with strings as keys instead of the Compound.Properties

        Args:
            force_classification (bool, optional): whether to force all classification.
            Defaults to True.

        Returns:
            dict[str, int]: formatted classification dict
        """
        if force_classification:
            self.classify()
        return {
            Compound.classification_keys[key]: value
            for key, value in self._classification.items()
        }



"""Implement Ortho classification.

Conains the code that manages the classification of ortho effects over
all the compound, transformation and circle level of the classification
workflow of the Nonadditivity Analysis.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from rdkit import Chem

from nonadditivity.classification.transformation_classification import is_h_replaced
from nonadditivity.classification.utils import list_empty

if TYPE_CHECKING:
    from nonadditivity.classification import Compound, Transformation
    from nonadditivity.utils.types import Molecule


# ===========================================================
#
#           COMPOUND ORTHO FUNCTIONS
#
# ===========================================================


def get_correct_ring_match(
    whole_match: tuple[int],
    ring_matches: tuple[tuple[int], ...],
) -> set[int] | None:
    """Return the ring match that is included in whole match.

    Args:
        whole_match (set[int]): set to match ring to
        ring_matches (tuple[tuple[int]]): list of ring matches

    Returns:
        set[int] | None: correct ring match
    """
    whole_set = set(whole_match)
    ring_size = len(whole_set) - 3
    for ring_match in ring_matches:
        ring_match_set = set(ring_match)
        if len(whole_set.intersection(ring_match_set)) == ring_size:
            return ring_match_set
    return None


def match_ortho_patterns(
    molecule: Molecule,
) -> list[tuple[set[int], ...]]:
    """Find Ortho substituted aromatic rings in a molecule.

    Takes a molecule and matches the following to smarts:
    '[A,a]a1a([A,a][A,a])aaaa1', '[A,a]a1a([A,a][A,a])aaa1'.
    It returns tuples containing the matches as well as the
    matches for only the ring part of the patterns.

    Args:
        molecule (Molecule): molecule to match ortho patterns

    Returns:
        list[tuple[set[int], Optional[set[int]]]]: list of tuples
        containing the ortho match as well as only the ring part
        that is part of the ortho match.
    """
    ortho_patterns = [
        (
            Chem.MolFromSmarts(patt[0]),  # type: ignore pylint: disable=E1101
            Chem.MolFromSmarts(patt[1]),  # type: ignore pylint: disable=E1101
        )
        for patt in [
            ("[A,a]a1a([A,a][A,a])aaaa1", "a1aaaaa1"),
            ("[A,a]a1a([A,a][A,a])aaa1", "a1aaaa1"),
        ]
    ]
    matches = []
    for pattern, ring_pattern in ortho_patterns:
        whole_matches = molecule.GetSubstructMatches(pattern)
        ring_matches = molecule.GetSubstructMatches(ring_pattern)
        for match in whole_matches:
            ring_match = get_correct_ring_match(
                whole_match=match,
                ring_matches=ring_matches,
            )
            if ring_match is None:
                continue
            matches.append((set(match), ring_match))
    return matches


def get_num_ortho_configurations(**kwargs: Molecule) -> int:
    """Get number of ortho matches.

    Returns number of aromatic rings with ortho substituents
    in a molecule. Can return up to 2n of n ortho configurations
    since some of them are matched twice.

    use keyword argument 'molecule' for it to work.

    Returns:
        int: Num of ortho substituents in molecule.
    """
    return len(get_ortho_indices(**kwargs))


def get_ortho_indices(**kwargs: Molecule) -> list[tuple[set[int], ...]]:
    """Get indices of possible ortho substituents.

    Gets the indices of atoms in a molecule that are
    part of ortho substituted aromatic rings, including
    the rings and 1, respectively 2 atoms of the substituents.
    together with the whole match it returns the indices of just
    the ring that is matched in each ortho match.

    use keyword argument 'molecule' for it to work.

    Returns:
        list[tuple[set[int],set[int]]]: tuple of (ortho_match, ring_match)
    """
    molecule: Molecule = kwargs["molecule"]
    return match_ortho_patterns(molecule)


# ===========================================================
#
#           TRANSFORMATION ORTHO FUNCTIONS
#
# ===========================================================


def differentiate_r_groups(
    non_ring_indices: set[int],
    compound: Compound,
) -> tuple[set[int], int | None]:
    """Differentiat indices of bigger and smaller r group.

    Returns tuple of a set of the indices that are in the bigger r group
    and the single index of the atom in the smaller r group.

    Args:
        non_ring_indices (set[int]): set of indices not in the aromatic
        ring investigated.
        compound (Compound): compound under investigation

    Returns:
        tuple[set[int], int | None]: set[indices longer r grouop],
        index shorter r group
    """
    assert len(non_ring_indices) == 3
    same_side = set()
    lonely_side = None
    for non_ring_index in non_ring_indices:
        neighbors = {
            atom.GetIdx()
            for atom in compound.rdkit_molecule.GetAtomWithIdx(
                non_ring_index,
            ).GetNeighbors()
        }
        if neighbors.intersection(non_ring_indices):
            same_side.add(non_ring_index)
            continue
        lonely_side = non_ring_index
    return same_side, lonely_side


def substituent_indices_in_same_ring(
    non_ring_indices: set[int],
    compound: Compound,
) -> bool:
    """Check whether both substituents are in same ring.

    Checks wheter substituents R1 and R2 in a pattern like R1-*-*-R2 are
    in the same ring.

    Args:
        non_ring_indices (set[int]): indices of R1 and R2
        compound (Compound): the compound in which this occurs.

    Returns:
        bool: True if R1 and R2 are in the same ring.
    """
    same_side, lonely_side = differentiate_r_groups(
        non_ring_indices=non_ring_indices,
        compound=compound,
    )
    ring_info = compound.rdkit_molecule.GetRingInfo()
    sameside = False
    for same in same_side:
        if ring_info.AreAtomsInSameRing(same, lonely_side):
            sameside = True
    return sameside


def _ortho_is_bicycle(
    non_ring_indices: set[int],
    ring_indices: set[int],
    compound: Compound,
) -> bool:
    """Check whether ortho pattern is a bicycle.

    Returns true if the longer substituent is part
    of a bicycle with the aromatic ring.

    Args:
        non_ring_indices (set[int]): indices not in ring
        ring_indices (set[int]): ring indices
        compound (Compound): compound object

    Returns:
        bool: true if bicycle
    """
    same_side, _ = differentiate_r_groups(
        non_ring_indices=non_ring_indices,
        compound=compound,
    )
    neighbors = [
        atom.GetIdx()
        for index in same_side
        for atom in compound.rdkit_molecule.GetAtomWithIdx(
            index,
        ).GetNeighbors()
    ]
    anchor = ring_indices.intersection(set(neighbors)).pop()
    assert isinstance(anchor, int)
    ring_info = compound.rdkit_molecule.GetRingInfo()
    return any(ring_info.AreAtomsInSameRing(s, anchor) for s in same_side)


def _atoms_are_neighbors(compound: Compound, indices: set[int]) -> bool:
    """Check whether atoms are neighbours.

    Returns if two atoms with indices given in 'indices' are neighbors in a
    compound.

    Args:
        compound (Compound): Compound
        indices (set[int]): 2 atom indices

    Returns:
        bool: True if atoms are connected via 1 bond.
    """
    assert len(indices) == 2
    return indices.pop() in [
        atom.GetIdx()
        for atom in compound.rdkit_molecule.GetAtomWithIdx(
            indices.pop(),
        ).GetNeighbors()
    ]


def _is_transformation_at_ortho(
    compound: Compound,
    constant_indices: list[int] | list[list[int]],
) -> int:
    """Check whether transformation is at ortho.

    Checks wheter a transformation is happening at the ortho position
    in an aromatic ring.

    Args:
        compound (Compound): compound to get ortho indices from
        constant_indices (list[int] | list[list[int]]): constant ondices of compound

    Returns:
        int: 1 == yes
             0 == no
              -1 == not identifiable (due to ambiguosly defined constant_indices)
    """
    ortho_match_found = False
    if list_empty(constant_indices) or isinstance(constant_indices[0], list):
        return -1
    ortho_matches = compound.get_property(compound.Properties.ORTHO_INDICES)
    assert isinstance(ortho_matches, list)
    constant_set = set(constant_indices)
    for ortho_match, ring_match in ortho_matches:
        ring_size = len(ring_match)
        # Check that ring is in constant part
        if len(constant_set.intersection(ring_match)) != ring_size:
            continue
        non_ring_indices = ortho_match.symmetric_difference(ring_match)
        const_substituents_intersect = constant_set.intersection(
            non_ring_indices,
        )
        # Check that longer substituent is also part of constant part
        if len(const_substituents_intersect) != 2:
            continue
        # Check that not one of either r group is matched with constant part.
        if not _atoms_are_neighbors(
            compound=compound,
            indices=const_substituents_intersect,  # type:ignore
        ):
            continue
        # Check that the two r groups are not in the same ring
        if substituent_indices_in_same_ring(
            non_ring_indices=non_ring_indices,
            compound=compound,
        ):
            continue
        # Check that the longer substituent is not in a bisycle with the ring match.
        if _ortho_is_bicycle(
            non_ring_indices=non_ring_indices,
            ring_indices=ring_match,
            compound=compound,
        ):
            continue
        ortho_match_found = True
    return int(ortho_match_found)


def transformation_at_ortho(**kwargs) -> int:
    """Check whether transformation happens at ortho.

    Looks at both sides of a transformation and tries to identify
    whether the transformation happens at the ortho position of an
    aromatic ring.

    provide keyword arguments 'compounds' and 'constant_indices'
    for this function to work.

    Returns:
        int: 1 == yes
             0 == no
              -1 == not identifiable (due to ambiguosly defined constant_indices)
    """
    ret = 0
    for compound, constant in zip(
        kwargs["compounds"],
        kwargs["constant_indices"],
    ):
        is_at_ortho = _is_transformation_at_ortho(
            compound=compound,
            constant_indices=constant,
        )
        if is_at_ortho == 1:
            return 1
        if is_at_ortho == -1:
            ret = -1
    return ret


def ortho_substituent_introduced(**kwargs) -> int:
    """Check whether ortho substituent is introduced.

    Determines wheter a substituent is introdced/removed at an ortho
    position of an aromatic ring in the molecule.

    provide keyword arguments 'compounds' and 'constant_indices' for
    this function to work

    Returns:
        int: 1 == yes
             0 == no
              -1 == not identifiable (due to ambiguosly defined constant_indices)
    """
    ortho = transformation_at_ortho(**kwargs)
    if ortho < 1:
        return ortho
    return int(is_h_replaced(**kwargs))


def ortho_substituent_exchanged(**kwargs) -> int:
    r"""Check whether ortho substituent is exchanged.

    Determines wheter a substituent is exchanged at an ortho position
    of an aromatic ring in the molecule

    provide keyword arguments 'compounds' and 'constant_indices' for
    this function to work.

    Returns:
        int: 1 == yes
            0 == no
             -1 == not identifiable (due to ambiguosly defined constant_indices)
    """
    ortho = transformation_at_ortho(**kwargs)
    if ortho < 1:
        return ortho
    return not int(is_h_replaced(**kwargs))


# ===========================================================
#
#           CIRCLE ORTHO FUNCTIONS
#
# ===========================================================


def get_ortho_transformation_changing(**kwargs: Transformation) -> bool:
    """Check whether circle has transformation at ortho position.

    Checks all transformation in a circle on wheter any of the transformations
    happens at the ortho position of an aromatic ring.

    Returns:
        bool: True if transformation at ortho position.
    """
    for i in range(4):
        transformation = kwargs[f"transformation_{i+1}"]
        if (
            transformation.get_property(
                transformation.Properties.ORTHO_SUBSTITUENT_CHANGES,
            )
            == 1
        ):
            return True
    return False


def get_ortho_transformation_introduced(**kwargs: Transformation) -> bool:
    """Check whether circle has introduction of ortho substituent.

    Checks all transformation in a circle on wheter any of the transformations
    happens at the ortho position of an aromatic ring.

    Returns:
        bool: True if transformation at ortho position.
    """
    for i in range(4):
        transformation = kwargs[f"transformation_{i+1}"]
        if (
            transformation.get_property(
                transformation.Properties.ORTHO_SUBSTITUENT_INTRODUCED,
            )
            == 1
        ):
            return True
    return False



"""Handle output writing for the NonadditivityAnalysis package."""

from pathlib import Path

import numpy as np
import pandas as pd

from nonadditivity.utils.log import get_logger
from nonadditivity.utils.math import mad_std, sn_medmed_std
from nonadditivity.workflow.canonicalize_output import (
    canonicalize_na_dataframe,
    get_transformation_df,
)

logger = get_logger()


def write_smiles_id_file(
    per_compound_dataframe: pd.DataFrame,
    smifile: Path,
) -> None:
    """Write input file for mmpdb fragment.

    Contains smiles and compoundid values.

    Args:
        per_compound_dataframe (pd.DataFrame): Dataframe to take the smiles from
        smifile (Union[Path, str]): path where to store the smiles.
    """
    per_compound_dataframe.to_csv(
        path_or_buf=smifile,
        sep="\t",
        columns=["SMILES", "Compound_ID"],
        index=False,
        header=False,
    )


def _write_c2c_file(
    c2c_file_path: Path,
    c2c_dataframe: pd.DataFrame,
) -> None:
    """Write file mapping circle to compounds.

    Args:
        c2c_file_path (str): path to write the file to.
        c2c_dataframe (pd.DataFrame): C2C dataframe to be written out.
    """
    c2c_dataframe.to_csv(
        path_or_buf=c2c_file_path,
        index=False,
    )


def _write_naa_file(path: Path, na_dataframe: pd.DataFrame) -> None:
    """Write the main nonadditivity output file.

    Args:
        path (str): path to file
        na_dataframe (pd.DataFrame): naa circle dataframe.
    """
    na_dataframe.to_csv(path_or_buf=path, index=False)


def _transform_per_compound_dataframe(
    per_compound_dataframe: pd.DataFrame,
    property_columns: list[str],
    include_censored: bool,
    series_column: str | None = None,
) -> pd.DataFrame:
    """Transform per compound dataframe to have a row for every property.

    Transforms the per_compound_dataframe in a way that every row only contains the
    information about one compound and one property respectively instead of one
    compound and mult. properties. It furhter removes censored values if
    include_censored == False.

    Args:
        per_compound_dataframe (pd.DataFrame): dataframe to transform
        property_columns (list[str]): name of the property columns
        include_censored (bool): whether to include censored values
        series_column (str | None, optional): Name of the series column. Defaults to
        None.

    Returns:
        pd.DataFrame: pivoted per compound data
    """
    columns = (
        [
            "Censors",
            "Nonadditivity_Pure",
            "Nonadditivity_Pure_SD",
            "Nonadditivity_Pure_Count",
            "Nonadditivity_Mixed",
            "Nonadditivity_Mixed_SD",
            "Nonadditivity_Mixed_Count",
        ]
        if series_column
        else [
            "Censors",
            "Nonadditivity",
            "Nonadditivity_SD",
            "Nonadditivity_Count",
        ]
    )

    new_pc_dataframe = pd.DataFrame()
    for column in property_columns:
        not_now = property_columns.copy()
        not_now.remove(column)
        new_df = per_compound_dataframe.drop(columns=not_now)
        new_df = new_df.dropna(subset=[column])
        new_df["Property"] = [column for _ in range(len(new_df))]
        new_df["Property_Value"] = list(new_df[column].values)
        new_df = new_df.drop(columns=column)
        for col in columns:
            new_df[col] = list(new_df[f"{column}_{col}"].values)
        to_drop = set()
        for index, censor in zip(
            new_df.index.values,
            new_df[f"{column}_Censors"].values,
        ):
            if censor == "NA":
                to_drop.add(index)
                continue
            if not include_censored and censor:
                to_drop.add(index)
        new_df = new_df.drop(index=list(to_drop))
        new_pc_dataframe = pd.concat([new_pc_dataframe, new_df])
    for column in property_columns:
        new_pc_dataframe = new_pc_dataframe.drop(
            columns=[f"{column}_{col}" for col in columns],
        )
    to_drop = set()

    return new_pc_dataframe


def _write_per_compound_file(
    path: Path,
    per_compound_dataframe: pd.DataFrame,
    property_columns: list[str],
    include_censored: bool,
    series_column: str | None = None,
) -> None:
    """Write per compound file contianing, compound, nonadditivities.

    Args:
        path (str): path to file
        per_compound_dataframe (pd.DataFrame): per compound dataframe.
        property_columns (list[str]): list containint prop column names.
        include_censored (bool): Wheter tho remove rows where censors present.
        series_column (str | None, optional): name of series column. defaults to None.
    """
    out_dataframe = _transform_per_compound_dataframe(
        per_compound_dataframe=per_compound_dataframe,
        property_columns=property_columns,
        include_censored=include_censored,
        series_column=series_column,
    )
    additional = (
        ["fused_ring_indices", "aromatic_indices"]
        if "fused_ring_indices" in out_dataframe.columns
        else []
    )
    out_dataframe = out_dataframe.drop(
        columns=["Neighbor_dict", "RDKit_Molecules"]
        + additional
        + [f"{col}_Nonadditivities" for col in property_columns],
    )
    out_dataframe = out_dataframe.rename({"Censors": "Property_Censors"})
    out_dataframe.to_csv(path_or_buf=path, index=False)


def _write_std_to_cmdline(
    property_columns: list[str],
    na_dataframe: pd.DataFrame,
    series_column: str | None = None,
) -> None:
    """Write calculated nonadditivities and std devs to the command line.

    Args:
        property_columns (list[str]): Name of the property columns in dataframe
        na_dataframe (pd.DataFrame): na_dataframe object
        series_column (str | None, optional): Name of the series_column. Defaults to
        None.
    """

    def create_logging_message(
        dataframe: pd.DataFrame,
        property_column: str,
        series: str | None = None,
    ) -> str:
        values = dataframe["Nonadditivity"].to_numpy()
        num_values = len(values)
        if num_values < 2:
            return (
                "There were not sufficient nonadditivity circles found "
                "to make a prediction on assay uncertainty "
                f"for property '{property_column}'!"
            )
        msg = (
            f"\n\nEstimated Experimental Uncertainty for property '{property_column}' "
        )
        if series:
            msg += f"and series '{series}' "
        msg += f"based on {num_values} nonadditivity circles:\n"
        msg += f"normal standard deviation: {.5* np.std(values)}\n"  # type: ignore
        msg += f"MAD standard deviation: {.5*mad_std(values)}\n"  # type: ignore
        msg += "Median of Medians standard deviation: "
        msg += str(0.5 * sn_medmed_std(values)) + "\n\n"  # type: ignore
        return msg

    if series_column:
        for property_column in property_columns:
            for series in np.unique(na_dataframe.Series.values):  # type: ignore
                out_df = na_dataframe[na_dataframe.Series == series]
                out_df = out_df[out_df.Property == property_column]
                logger.info(
                    create_logging_message(
                        dataframe=out_df,
                        property_column=property_column,
                        series=series,
                    ),
                )
        return
    for property_column in property_columns:
        out_df = na_dataframe[na_dataframe.Property == property_column]
        logger.info(
            create_logging_message(
                dataframe=out_df,
                property_column=property_column,
                series=None,
            ),
        )


def write_canonical_output(
    canonical_na_df: pd.DataFrame,
    canonical_tr_df: pd.DataFrame,
    naa_file_path: Path,
) -> None:
    """Write canonical dataframes to csv files.

    Args:
        canonical_na_df (pd.DataFrame): dataframe containing canonical circles
        canonical_tr_df (pd.DataFrame): dataframe containing canonical transformations
        naa_file_path (Path): file path for naa output dataframe.
    """
    _write_naa_file(
        path=naa_file_path.parent / "canonical_na_output.csv",
        na_dataframe=canonical_na_df,
    )
    canonical_tr_df.to_csv(
        path_or_buf=naa_file_path.parent / "canonical_transformations.csv",
        index=False,
    )


def write_output_files(
    per_compound_dataframe: pd.DataFrame,
    na_dataframe: pd.DataFrame,
    c2c_dataframe: pd.DataFrame,
    c2c_file_path: Path,
    naa_file_path: Path,
    per_compound_path: Path,
    property_columns: list[str],
    include_censored: bool,
    canonicalize: bool,
    series_column: str | None = None,
) -> None:
    """Write naa, c2c and pcp output files.

    Writes all the output files that are then needed for the spotfire
    representation of the data.

    Args:
        per_compound_dataframe (pd.DataFrame): per compound dataframe
        na_dataframe (pd.DataFrame): nonadditivity circles dataframe
        c2c_dataframe (pd.DataFrame): dataframe mapping circles to compounds
        c2c_file_path (str): path to file mapping circles and compounds
        naa_file_path (str): path to file containing nonadditivity circles
        per_compound_path (str): path to file containing per compound info
        include_censored (bool): Whether to include censored values.
        series_column (str | None, optional): series column if provided.
        property_columns (list[str]): name of the property columns used in analysis.
        canonicalize (bool): Whether to canonicalize input
    """
    _write_c2c_file(
        c2c_file_path=c2c_file_path,
        c2c_dataframe=c2c_dataframe,
    )
    _write_naa_file(path=naa_file_path, na_dataframe=na_dataframe)
    _write_per_compound_file(
        path=per_compound_path,
        per_compound_dataframe=per_compound_dataframe,
        property_columns=property_columns,
        include_censored=include_censored,
        series_column=series_column,
    )
    _write_std_to_cmdline(
        property_columns=property_columns,
        na_dataframe=na_dataframe,
        series_column=series_column,
    )

    if canonicalize:
        canonical_na_df = canonicalize_na_dataframe(na_dataframe=na_dataframe)
        canonical_tr_df = get_transformation_df(canonical_df=canonical_na_df)
        write_canonical_output(
            canonical_na_df=canonical_na_df,
            canonical_tr_df=canonical_tr_df,
            naa_file_path=naa_file_path,
        )



"""Python wrapper for the mmpdblib code that is used in this package."""

import shlex
import subprocess
from collections.abc import Generator
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Literal

import pandas as pd

from nonadditivity.utils.errors import FragmentationError, IndexingError
from nonadditivity.utils.log import get_logger

logger = get_logger()


def _run_mmpdb_command(command: str) -> int:
    """Run command in shell and pipe output to logger.

    Wrapper for the mmpdb commandline interface so it would run the
    commands using subprocess while streaming its output to the logger
    in realtime.

    Args:
        command (str): command to execute, will be formatted with shlex
        logger (logging.Logger): logger to write messages to.

    Returns:
        int: return code of the application.
    """

    def run_subprocess(
        command: str,
    ) -> Generator[tuple[str, int | None] | tuple[Literal[""], int], Any, None]:
        process = subprocess.Popen(
            shlex.split(command),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        stdout = process.stdout.fileno()  # type: ignore
        with open(stdout, encoding="utf-8") as nice_stdout:
            for line in nice_stdout:
                yield line.replace("\n", " "), process.poll()

            yield "", process.wait()

    for line, returncode in run_subprocess(command=command):
        if returncode is not None:
            return returncode
        if line.isspace():
            continue
        logger.info(line)
    return 0


def run_mmpdb_fragment(
    cache: bool,
    fragments_file: Path,
    structure_file: Path,
) -> None:
    """Run `mmpdb fragment` command.

    Args:
        cache (bool): whether to use old fragmentation
        fragments_file (Path): path to fragmentation file
        structure_file (Path): path to input structure file
    """
    command = (
        f"mmpdb fragment --num-jobs {cpu_count() -2 } --delimiter tab "
        f"--output {fragments_file.as_posix()} {structure_file.as_posix()} "
    )
    if cache:
        command += f"--cache {fragments_file.as_posix()}"
    else:
        command += "-i smi --max-rotatable-bonds 20"

    if _run_mmpdb_command(command=command) != 0:
        raise FragmentationError(
            "Something happened during mmpdb fragmentation. Please check the logs!",
        )


def run_mmpdb_index(
    fragments_file: Path,
    mmp_outputfile: Path,
    max_variable_heavies: int,
) -> None:
    """Run `mmpdb index` command.

    Args:
        fragments_file (Path): fragmentation file
        mmp_outputfile (Path): output file
        max_variable_heavies (int): max heavy atom count to be exchanged in
        a mmp.
    """
    command = (
        f"mmpdb index --out csv --symmetric --output {mmp_outputfile.as_posix()} "
        f"--max-variable-heavies {max_variable_heavies} {fragments_file.as_posix()}"
    )
    if _run_mmpdb_command(command=command) != 0:
        raise IndexingError(
            "Something happened during mmpdb indexing. Please check the logs!",
        )


def read_raw_mmps(path: Path) -> pd.DataFrame:
    """Read in file generated by `mmpdb index`.

    Args:
        path (str): path to the file to be read in.

    Returns:
        pd.DataFrame: file content in a pandas dataframe.
    """
    mmp_dataframe = pd.read_table(
        filepath_or_buffer=path,
        sep="\t",
        header=None,
        names=[
            "SMILES_LHS",
            "SMILES_RHS",
            "ID_LHS",
            "ID_RHS",
            "TRANSFORMATION",
            "CONSTANT",
        ],
    )
    return mmp_dataframe


def run_mmpdlib_code(
    cache: bool,
    fragments_file: Path,
    mmp_outputfile: Path,
    structure_file: Path,
    max_variable_heavies: int,
) -> pd.DataFrame:
    """Run mmpdb fragment and index.

    Runs all the mmpdblib code in one and returns the matched molecular pairs as a list
    of strings. The strings contain SMILES and IDs of the whole molecules as well as
    fragment SMILES

    Args:
        cache (bool): wheter you want to use previously fragmented molecules.
        fragments_file (str): path to where the fragments are stored and
        where previously fragmented molecules would be stored.
        mmp_outputfile (str): path to write the matched molecular pair data.
        structure_file (str): input file containing smiles of the molecules
        you want to find mmps for.
        max_variable_heavies (int): max heavy atoms to be replaced in an mmp.

    Returns:
        DataFrame: Pandas Dataframe containing the mmpdb results.
    """
    if cache and not fragments_file.exists():
        logger.warning(
            "Was not able to locate results from previous fragmentation.",
        )
        logger.warning("Will redo all fragmentation.")
        cache = False

    run_mmpdb_fragment(
        cache=cache,
        fragments_file=fragments_file,
        structure_file=structure_file,
    )
    run_mmpdb_index(
        fragments_file=fragments_file,
        mmp_outputfile=mmp_outputfile,
        max_variable_heavies=max_variable_heavies,
    )
    return read_raw_mmps(path=mmp_outputfile)



"""Workflows for nonadditivity analysis package."""

from nonadditivity.workflow.input_parsing import parse_input_file
from nonadditivity.workflow.mmpdb_helper import run_mmpdlib_code
from nonadditivity.workflow.nonadditivity_core import run_nonadditivity_core
from nonadditivity.workflow.output import write_output_files, write_smiles_id_file



"""Impolements the core algorithm for the nonadditivity analysis."""

import random
from typing import Any

import pandas as pd
from rdkit import Chem
from tqdm.auto import tqdm

from nonadditivity.utils.log import get_logger
from nonadditivity.utils.math import (
    _calculate_mean_and_std_per_compound,
    _calculate_nonadditivity,
    _calculate_theo_quantiles,
)

logger = get_logger()

sign_switch: dict[int, int] = {0: 1, 1: -1}


def top_distance_changes(var_lhs: str, var_rhs: str) -> bool:
    """Check whether a linker exchange is valid.

    Checks whether the topological distance between two or three anchoring points
    is the same if a linker is exchanged.

    Args:
        var_lhs (str): SMILES of compound 1
        var_rhs (str): SMILES of compound 2

    Returns:
        bool: False if distance is the same.
    """
    fragment_a = Chem.MolFromSmarts(var_lhs)  # type: ignore pylint: disable=E1101
    fragment_b = Chem.MolFromSmarts(var_rhs)  # type: ignore pylint: disable=E1101
    atoms_a = [atom.GetSmarts() for atom in fragment_a.GetAtoms()]
    atoms_b = [atom.GetSmarts() for atom in fragment_b.GetAtoms()]
    matrix_a = Chem.GetDistanceMatrix(fragment_a)  # type: ignore pylint: disable=E1101
    matrix_b = Chem.GetDistanceMatrix(fragment_b)  # type: ignore pylint: disable=E1101

    a_idx1 = atoms_a.index("[*:1]")
    a_idx2 = atoms_a.index("[*:2]")
    b_idx1 = atoms_b.index("[*:1]")
    b_idx2 = atoms_b.index("[*:2]")
    if matrix_a[a_idx1, a_idx2] != matrix_b[b_idx1, b_idx2]:
        return True
    if "[*:3]" in var_lhs:
        a_idx3 = atoms_a.index("[*:3]")
        b_idx3 = atoms_b.index("[*:3]")
        if matrix_a[a_idx1, a_idx3] != matrix_b[b_idx1, b_idx3]:
            return True
        if matrix_a[a_idx2, a_idx3] != matrix_b[b_idx2, b_idx3]:
            return True
    return False


def add_neighbor_dictionary_to_df(
    mmp_dataframe: pd.DataFrame,
    per_compound_dataframe: pd.DataFrame,
    no_chiral: bool,
    verbose: bool = True,
) -> pd.DataFrame:
    """Get Shortest transformation from mmpdb output.

    Checks the data generated by mmpdblib index, and adds a column to the
    v giving neighbors and the respective transformation and constant part
    for every entry. All done inplace.

    !We could also run mmpdb with --smallest-transformation-only, but this
    is not faster than doing it here. As of November 2022!

    Args:
        mmp_dataframe (pd.DataFrame): data generated by mmpdblib
        per_compound_dataframe (pd.DataFrame): dataframe to add the column with
        neigbors to.
        no_chiral (bool, optional): True if you want to skip chiral compounds. Defaults
        to False.
        verbose (bool): wheter to display tqdm status bar. Defaults to True.
    """
    neigh_dict = {key: {} for key in per_compound_dataframe.index.to_numpy()}
    const_dict = {key: {} for key in per_compound_dataframe.index.to_numpy()}
    seen = {key: set() for key in per_compound_dataframe.index.to_numpy()}
    lengths = {key: {} for key in per_compound_dataframe.index.to_numpy()}
    for id_lhs, id_rhs, transf, constant in tqdm(
        zip(
            mmp_dataframe["ID_LHS"].values,
            mmp_dataframe["ID_RHS"].values,
            mmp_dataframe["TRANSFORMATION"].values,
            mmp_dataframe["CONSTANT"].values,
        ),
        desc="Analyzing Neighborhoods",
        total=len(mmp_dataframe["ID_LHS"].values),
        ascii="░▒█",
        disable=not verbose,
    ):
        if no_chiral and "@" in transf:
            continue
        var_lhs, var_rhs = transf.split(">>")
        # Skip pair if the transformation has more than one anchoring point
        # and the topological distance changes between those two (no
        # reason to assume additivity then)
        if "[*:2]" in var_lhs and top_distance_changes(
            var_lhs=var_lhs,
            var_rhs=var_rhs,
        ):
            continue
        newlen = len(transf)
        if id_rhs not in seen[id_lhs]:
            seen[id_lhs].add(id_rhs)
            neigh_dict[id_lhs][id_rhs] = transf
            const_dict[id_lhs][id_rhs] = constant
            lengths[id_lhs][id_rhs] = newlen
            continue

        # Add to neighbor dictionary
        if newlen < lengths[id_lhs][id_rhs]:
            neigh_dict[id_lhs][id_rhs] = transf
            const_dict[id_lhs][id_rhs] = constant
            lengths[id_lhs][id_rhs] = newlen

    for key, value in neigh_dict.items():
        for n_key, n_value in value.items():
            neigh_dict[key][n_key] = (n_value, const_dict[key][n_key])
    per_compound_dataframe["Neighbor_dict"] = neigh_dict.values()
    return per_compound_dataframe


def _get_tranformation_pairs(
    neighbordict: dict[str, dict[str, str]],
) -> dict[str, list[tuple[str, ...]]]:
    """Create dict mapping id pair to transformation.

    Creates dictionary that maps a transformation to a list with
    all compound pairs (ID1, ID2) with the relation Transformation(ID1) -> ID2
    only returns transformation that have more than one entry.

    Args:
        neighbordict (dict[str, dict[str, str]]): dictionary with IDs\
            , their neighbors and transformations.

    Returns:
        dict[str, list[tuple[str, ...]]]: {Transformation: [(ID1, ID2),...]}
    """
    transformation_pairs: dict[str, list[tuple[str, ...]]] = {}
    for compound_id, neighbor_dictionary in neighbordict.items():
        for neighbor_id, transformation in neighbor_dictionary.items():
            transformation_pairs[transformation] = transformation_pairs.get(
                transformation,
                [],
            )
            transformation_pairs[transformation].append(
                (compound_id, neighbor_id),
            )
    return transformation_pairs


def _find_circles(
    neighbordict: dict[str, dict[str, str]],
    transformation_pairs: dict[str, list[tuple[str, ...]]],
    verbose: bool = True,
) -> list[list[str]]:
    r"""Find Double transformation cycles.

    Core algorithm of the nonadditivity package.
    Following Name convention is followed:

    compound_1 --------t1--------> compound_2
        |                            |
        t2                           t3
        \/                           \/
    compound_4 --------t4-------> compound_3

    entries are appended to return list as [compound_1, compound_2, compound_3,
    compound_4]

    Args:
        neighbordict (dict[str, dict[str, str]]): dict mapping compound id to
        a dict with its neighbors and the transformations to them.
        transformation_pairs (dict[str, list[tuple[str, ...]]]): dict mapping
        transformation to tuples of comoound ids that can be mutated into each other by
        the transformation.
        verbose (bool): wheter to display tqdm status bar. Defaults to True.

    Returns:
        list[list[int]]: list of circle lists[compound_1, compound_2, compound_3,
        compound_4]
    """
    already_checked_1 = {key: False for key in neighbordict}
    already_checked_2 = {key: False for key in neighbordict}
    circles = []
    for compound_1, compound_1_neighbors in tqdm(
        neighbordict.items(),
        desc="Finding Double Transformation Cycles",
        ascii="░▒█",
        disable=not verbose,
    ):
        # No circle possible if only one neighbor to compound_1!
        if len(compound_1_neighbors.values()) > 1:
            already_checked_2 = already_checked_1.copy()
            for (
                compound_2,
                compound_1_to_compound_2,
            ) in compound_1_neighbors.items():
                compound_2_neighbors = neighbordict[compound_2]
                if already_checked_1[compound_2]:
                    continue
                for compound_4, compound_3 in transformation_pairs[
                    compound_1_to_compound_2
                ]:
                    if any(
                        [
                            compound_4 not in compound_1_neighbors,
                            compound_3 not in compound_2_neighbors,
                            already_checked_2[compound_4],
                            already_checked_1[compound_3],
                            compound_2 == compound_4,
                            compound_1 == compound_3,
                        ],
                    ):
                        continue
                    if (
                        compound_1_neighbors[compound_4]
                        == compound_2_neighbors[compound_3]
                    ):
                        circles.append(
                            [compound_1, compound_2, compound_3, compound_4],
                        )
                already_checked_2[compound_2] = True
        already_checked_1[compound_1] = True
    return circles


def get_circles(
    per_compound_dataframe: pd.DataFrame,
    verbose: bool = True,
) -> tuple[list[list[str]], dict[str, list[tuple[str, ...]]]]:
    """Find all dtcs for the compounds in the per compound dataframe.

    Wrapper for the core algorithm of the nonadditivity package.
    Removes all

    Args:
        per_compound_dataframe (pd.DataFrame): per compound data
        verbose (bool): If true, more is logged.

    Returns:
        tuple[list[list[str]],  dict[str, list[tuple[str, ...]]]]: all
        double transformation circles as well as the corresponding trans
        formations.
    """
    per_compound_dataframe = per_compound_dataframe[
        per_compound_dataframe.Neighbor_dict != {}
    ]
    neighbordict = {
        key: {k: v[0] for k, v in value.items()}
        for key, value in zip(
            per_compound_dataframe.index.values,
            per_compound_dataframe.Neighbor_dict.values,
        )
    }
    transformation_pairs = _get_tranformation_pairs(neighbordict=neighbordict)
    circles = _find_circles(
        neighbordict=neighbordict,
        transformation_pairs=transformation_pairs,
        verbose=verbose,
    )
    return circles, transformation_pairs


def _randomize_circles(
    circles: list[list[str]],
) -> tuple[list[list[str]], list[int]]:
    """Randomize circles.

    Takes a list of circles and randomly chooses starting point for every circle
    and returns new list of then randomized circles.

    Args:
        circles (list[list[str]]): original circles

    Returns:
        tuple[list[list[str]], list[int]]: randomized circles, shift for every
        randomizer circle
    """
    new_circles = []
    min_indices = []
    for circle in circles:
        min_index = random.choice(range(4))
        new_circles.append((circle + circle)[min_index : min_index + 4])
        min_indices.append(min_index)
    return new_circles, min_indices


def _create_circle_ids(
    circles: list[list[str]],
    property_column: str,
) -> list[str]:
    """Create IDS for all circles.

    Args:
        circles (list[list[str]]): list of circles
        property_column (str): name of the property that is investigated.

    Returns:
        list[str]: list with circle ids.
    """
    return [
        "_".join([str(c[0]), str(c[1]), str(c[2]), str(c[3])]) + f"_{property_column}"
        for c in circles
    ]


def _is_same_series(
    per_compound_dataframe: pd.DataFrame,
    circle: list[str],
) -> bool:
    """True if all compounds of a circle are in the same series.

    Args:
        per_compound_dataframe (pd.DataFrame): per compound dataframe
        circle (list[str]): circle to be checked.

    Returns:
        bool: true if all compounds are in the same series.
    """
    return len({per_compound_dataframe.loc[srn, "Series"] for srn in circle}) == 1


def _get_faulty_entries(
    per_compound_dataframe: pd.DataFrame,
    circles: list[list[str]],
    property_column: str,
    include_censored: bool,
) -> set[str]:
    faulty = set()
    censor_column = f"{property_column}_Censors"
    for index, circle in enumerate(circles):
        for compound in circle:
            if per_compound_dataframe.loc[compound, censor_column] == "NA":
                faulty.add(index)
                continue
            if (
                not include_censored
                and per_compound_dataframe.loc[compound, censor_column]
            ):
                faulty.add(index)
    return faulty


def create_c2c_dataframe(c2c_vals: dict[str, list[Any]]) -> pd.DataFrame:
    """Create dataframe mapping circle to compounds.

    Args:
        c2c_vals (dict[str, list[Any]]): values to be added to new dataframe.

    Returns:
        pd.DataFrame: dataframe with c2c vals.
    """
    c2c_dataframe = pd.DataFrame()
    c2c_dataframe["Circle_ID"] = c2c_vals["circle_ids"]
    c2c_dataframe["Compound_ID"] = c2c_vals["compound_ids"]
    c2c_dataframe["SMILES"] = c2c_vals["smiles"]
    c2c_dataframe["Property"] = c2c_vals["properties"]
    c2c_dataframe["NonadditivityPerCompound"] = c2c_vals["nonadditivities"]
    return c2c_dataframe


def create_na_dataframe(
    na_vals: dict[str, list[Any]],
    randomized_circles: list[list[str]],
    per_compound_dataframe: pd.DataFrame,
) -> pd.DataFrame:
    """Create NAA dataframe containing circles and their nonadditivity.

    Args:
        na_vals (dict[str, list[Any]]): nonadditivity values
        randomized_circles (list[list[str]]): circle ids
        per_compound_dataframe (pd.DataFrame): per compound dataframe

    Returns:
        pd.DataFrame: nonadditivity data
    """
    na_dataframe = pd.DataFrame()
    for i in range(4):
        na_dataframe[f"Compound{i+1}"] = [circle[i] for circle in randomized_circles]
    for i in range(4):
        na_dataframe[f"SMILES{i+1}"] = [
            per_compound_dataframe.loc[circle[i], "SMILES"]
            for circle in randomized_circles
        ]
    for i in range(4):
        na_dataframe[f"Prop_Cpd{i+1}"] = na_vals[f"cpdvals{i}"]

    for i, v in zip([1, 2], [1, 3]):
        na_dataframe[f"Transformation{i}"] = [
            per_compound_dataframe.loc[circle[0], "Neighbor_dict"][circle[v]][0]
            for circle in randomized_circles
        ]
    for name in [
        "Property",
        "Series",
        "Nonadditivity",
        "Theo_Quantile",
        "Circle_ID",
    ]:
        na_dataframe[name] = na_vals[name]
    return na_dataframe


def calculate_na_output(
    circles: list[list[str]],
    per_compound_dataframe: pd.DataFrame,
    property_columns: list[str],
    include_censored: bool,
    series_column: str | None = None,
) -> tuple[pd.DataFrame, list[list[str]], list[str], pd.DataFrame]:
    """Create dataframe with nonadditivity circles.

    Creates a dataframe with an entry for every nonadditivity cicle and property.
    also returns randomized circles and their corresponding circle ids.

    Args:
        circles (list[list[str]]): list of nonadditivity circles
        per_compound_dataframe (pd.DataFrame): per compound dataframe
        property_columns (list[str]): name of investigated property columns
        include_censored (bool): Whether to also use censored values.
        series_column (str | None, optional): name of the series column. Defaults to
        None.

    Returns:
        tuple[pd.DataFrame, list[list[str]], list[str]]: nonadditivity dataframe,
        randomized circles, circle_ids
    """
    na_vals: dict[str, list[float] | list[str]] = {
        key: []
        for key in [
            "Circle_ID",
            "Property",
            "Nonadditivity",
        ]
        + [f"cpdvals{i}" for i in range(4)]
    }
    c2c_vals: dict[str, list[float] | list[str]] = {
        key: []
        for key in [
            "circle_ids",
            "compound_ids",
            "smiles",
            "properties",
            "nonadditivities",
        ]
    }

    randomized_circles = []

    for property_column in property_columns:
        indices_to_remove = _get_faulty_entries(
            per_compound_dataframe=per_compound_dataframe,
            circles=circles,
            property_column=property_column,
            include_censored=include_censored,
        )
        valid_circles = [c for i, c in enumerate(circles) if i not in indices_to_remove]
        if len(valid_circles) < 1:
            _update_na_per_compound(
                per_compound_dataframe=per_compound_dataframe,
                circle=[],
                nonadditivity=0,
                property_column=property_column,
                series_column=series_column,
            )
            continue
        rcircles, _ = _randomize_circles(circles=valid_circles)
        randomized_circles += rcircles
        na_vals["Circle_ID"] += _create_circle_ids(
            circles=rcircles,
            property_column=property_column,
        )
        na_vals["Property"] += [property_column for _ in rcircles]
        for i in range(4):
            na_vals[f"cpdvals{i}"] += [
                per_compound_dataframe.loc[circle[i], property_column]
                for circle in rcircles
            ]
        nonadditivity = [
            _calculate_nonadditivity(
                values=[
                    per_compound_dataframe.loc[index, property_column]
                    for index in circle
                ],
            )
            for circle in rcircles
        ]
        na_vals["Nonadditivity"] += nonadditivity

        for circle, non_add in zip(
            rcircles,
            nonadditivity,
        ):
            _update_na_per_compound(
                per_compound_dataframe=per_compound_dataframe,
                circle=circle,
                nonadditivity=non_add,
                property_column=property_column,
                series_column=series_column,
            )
        cmpds = [i for c in rcircles for i in c]

        c2c_vals["compound_ids"] += cmpds
        c2c_vals["smiles"] += [per_compound_dataframe.loc[i, "SMILES"] for i in cmpds]
        c2c_vals["nonadditivities"] += [
            sign_switch[i % 2] * na
            for circle, na in zip(rcircles, nonadditivity)
            for i, _ in enumerate(circle)
        ]

        c2c_vals["properties"] += [property_column for _ in cmpds]

    c2c_vals["circle_ids"] = [i for i in na_vals["Circle_ID"] for _ in range(4)]
    if series_column:
        na_vals["Series"] = [
            (
                per_compound_dataframe.loc[circle[0], "Series"]
                if _is_same_series(
                    per_compound_dataframe=per_compound_dataframe,
                    circle=circle,
                )
                else ""
            )
            for circle in randomized_circles
        ]
    else:
        na_vals["Series"] = ["" for _ in randomized_circles]

    na_vals["Theo_Quantile"] = _calculate_theo_quantiles(
        series_columns=series_column,
        series_ids=na_vals["Series"],  # type: ignore
        nonadditivities=na_vals["Nonadditivity"],
    )

    return (
        create_na_dataframe(
            na_vals=na_vals,
            randomized_circles=randomized_circles,
            per_compound_dataframe=per_compound_dataframe,
        ),
        randomized_circles,
        na_vals["Circle_ID"],
        create_c2c_dataframe(c2c_vals=c2c_vals),
    )


def _update_na_per_compound(
    per_compound_dataframe: pd.DataFrame,
    circle: list[str],
    nonadditivity: float,
    property_column: str,
    series_column: str | None = None,
) -> None:
    """Adds nonadditivit value to per compound dataframe.

    Args:
        per_compound_dataframe (pd.DataFrame): per compound dataframe.
        circle (list[str]): investigated circle
        nonadditivity (float): nonadditivity value for circle
        series_column (str | None, optional): name of series column. Defaults to None
        property_column (str | None, optional): investigated proeprty column name.
        Defaults to None
    """
    column = f"{property_column}_Nonadditivities"
    if column not in per_compound_dataframe.columns:
        per_compound_dataframe[column] = [
            [] for _ in range(len(per_compound_dataframe))
        ]
    for index, compound in enumerate(circle):
        if series_column:
            is_pure = (
                "pure" if _is_same_series(per_compound_dataframe, circle) else "mixed"
            )
            per_compound_dataframe.loc[compound, column].append(  # type: ignore
                (sign_switch[index % 2] * nonadditivity, is_pure),
            )
        else:
            per_compound_dataframe.loc[compound, column].append(  # type: ignore
                (sign_switch[index % 2] * nonadditivity, None),
            )


def add_mean_and_std_na_to_df(
    per_compound_dataframe: pd.DataFrame,
    property_columns: list[str],
    series_column: str | None = None,
) -> None:
    """Add mean, std and count to per compound dataframe.

    Creates mean, std, and count for all nonadditivities per compound in the per
    compound dataframe and adds new columns with mean, std, and num occurence.
    If series column is given, columns for the pure and the mixed values are added.

    Args:
        per_compound_dataframe (pd.DataFrame): per compound dataframe
        property_columns (list[str]): name of the property columns investigated
        series_column (str | None, optional): series column name. Defaults to None.
    """
    for property_column in property_columns:
        per_compound_values = _calculate_mean_and_std_per_compound(
            per_compound_dataframe=per_compound_dataframe,
            property_column=property_column,
            series_column=series_column,
        )
        if series_column:
            per_compound_dataframe[f"{property_column}_Nonadditivity_Pure"] = [
                value[0]
                for value in per_compound_values[0]  # type: ignore
            ]
            per_compound_dataframe[f"{property_column}_Nonadditivity_Pure_SD"] = [
                value[1]
                for value in per_compound_values[0]  # type: ignore
            ]
            per_compound_dataframe[f"{property_column}_Nonadditivity_Pure_Count"] = [
                value[2]
                for value in per_compound_values[0]  # type: ignore
            ]
            per_compound_dataframe[f"{property_column}_Nonadditivity_Mixed"] = [
                value[0]
                for value in per_compound_values[1]  # type: ignore
            ]
            per_compound_dataframe[f"{property_column}_Nonadditivity_Mixed_SD"] = [
                value[1]
                for value in per_compound_values[1]  # type: ignore
            ]
            per_compound_dataframe[f"{property_column}_Nonadditivity_Mixed_Count"] = [
                value[2]
                for value in per_compound_values[1]  # type: ignore
            ]
        else:
            per_compound_dataframe[f"{property_column}_Nonadditivity"] = [
                value[0]
                for value in per_compound_values  # type: ignore
            ]
            per_compound_dataframe[f"{property_column}_Nonadditivity_SD"] = [
                value[1]
                for value in per_compound_values  # type: ignore
            ]
            per_compound_dataframe[f"{property_column}_Nonadditivity_Count"] = [
                value[2]
                for value in per_compound_values  # type: ignore
            ]


def run_nonadditivity_core(
    mmp_dataframe: pd.DataFrame,
    per_compound_dataframe: pd.DataFrame,
    property_columns: list[str],
    no_chiral: bool,
    include_censored: bool,
    series_column: str | None = None,
    verbose: bool = True,
) -> tuple[pd.DataFrame, list[list[str]], list[str], pd.DataFrame]:
    """Executes the whole nonadditivity core functions.

    takes an mmp_dataframe and a per compound dataframe and finds
    nonaditivity circles, calculates their nonadditivity, and then
    calculats the nonadditivity per compound.

    Args:
        mmp_dataframe (pd.DataFrame): mmp dataframe from mmpdblib
        per_compound_dataframe (pd.DataFrame): per compound dataframe
        property_columns (list[str]): name of the property columns
        no_chiral (bool): If true, transformation containing @ are skipped.
        include_censored (bool): If true, also include censored values.
        series_column (str | None): name of the series column. Defaults to None.
        verbose (bool, optional): If true, more output is logged. Defaults to True.

    Returns:
        tuple[pd.DataFrame, list[list[str]], list[str]]: nonadditivity dataframe,
        randomized circles, circle_ids
    """
    per_compound_dataframe = add_neighbor_dictionary_to_df(
        mmp_dataframe=mmp_dataframe,
        per_compound_dataframe=per_compound_dataframe,
        no_chiral=no_chiral,
        verbose=verbose,
    )
    circles, _ = get_circles(
        per_compound_dataframe=per_compound_dataframe,
        verbose=verbose,
    )
    logger.info("%s Nonadditivity circles were found.", len(circles))
    (
        na_dataframe,
        randomized_circles,
        circle_ids,
        c2c_dataframe,
    ) = calculate_na_output(
        circles=circles,
        per_compound_dataframe=per_compound_dataframe,
        property_columns=property_columns,
        series_column=series_column,
        include_censored=include_censored,
    )
    add_mean_and_std_na_to_df(
        per_compound_dataframe,
        property_columns,
        series_column=series_column,
    )

    return na_dataframe, randomized_circles, circle_ids, c2c_dataframe



"""Handle commanline interface with click.

This file handles all the commandline input parsing for
the NonadditivityAnalysis package using click.

"""

import logging
import os
from collections.abc import Callable
from pathlib import Path

import click

VERBOSITY_LEVEL = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}


def add_entry_point_options(command: Callable) -> Callable:
    """Handle command line input.

    Adds all needed flags to an option
    that can then be run in the command line and parses them into an
    InputOptions file, that does the validity checks for the input.

    Args:
        command (Callable): function to add the command line options to.

    Returns:
        Callable: function with the command line options.
    """

    def set_click_attrs(dest: Callable, src: Callable) -> None:
        """Add __name__, __doc__ and __click_params__ from src to destination.

        Args:
            dest (Callable): function where src' attributes are copied to
            src (Callable): attributes from this function are copoied to dest
        """
        dest.__name__ = src.__name__
        dest.__doc__ = src.__doc__
        dest.__click_params__ = src.__click_params__  # type: ignore

    param_names = []

    def add_option(*args, **kwargs) -> None:
        """Add input click.options to function."""
        param_names.append(args[-1].lstrip("-").replace("-", "_"))
        click.option(*args, **kwargs)(command)

    add_option(
        "-i",
        "--infile",
        "infile_",
        help="Input file",
        type=click.Path(exists=True, resolve_path=True, path_type=Path),
        prompt="Path to the input .txt file",
        required=True,
    )

    add_option(
        "--update",
        "update_",
        is_flag=True,
        default=False,
        type=bool,
        help="Use fragmentation and images from previous run",
    )

    add_option(
        "-m",
        "--max-heavy",
        "max_heavy_",
        help="Maximum number of heavy atoms per ligand",
        type=int,
        default=70,
    )

    add_option(
        "--max-heavy-in-transformation",
        "max_heavy_in_transformation_",
        help=(
            "Maximum number of heavy atoms that are allowed to be "
            "exchanged in a MMP."
        ),
        type=int,
        default=10,
    )

    add_option(
        "--classify",
        "classify_",
        is_flag=True,
        default=False,
        type=bool,
        help=(
            "Set this flag if you want additional properties for your"
            " Nonadditivty cycles to be calculated."
        ),
    )

    add_option(
        "--canonicalize",
        "canonicalize_",
        is_flag=True,
        default=False,
        type=bool,
        help=(
            "Set this flag if you want additional output files where"
            " every tranformation is only occuring in one way and not."
            " not the other"
        ),
    )

    add_option(
        "--no-chiral",
        "no_chiral_",
        is_flag=True,
        default=False,
        type=bool,
        help="Skip all transformations that include chiral centers",
    )

    add_option(
        "-d",
        "--delimiter",
        "delimiter_",
        help="Specify delimiter in input file",
        type=click.Choice(["tab", "space", "comma", "semicolon"]),
        default="tab",
    )

    add_option(
        "--include-censored",
        "include_censored_",
        is_flag=True,
        default=False,
        type=bool,
        help=(
            "Include Circles where at least one out of four compounds has "
            "a censored value"
        ),
    )

    add_option(
        "-o",
        "--output-directory",
        "directory_",
        default=None,
        type=click.Path(writable=True, resolve_path=True, path_type=Path),
        help="Output directory name",
    )

    add_option(
        "--series-column",
        "series_column_",
        default=None,
        type=str,
        help="Column that identifies subseries within the dataset",
    )

    add_option(
        "-p",
        "--property-columns",
        "property_columns_",
        type=str,
        multiple=True,
        default=None,
        help="Property columns for which Nonadditivity should be calculated",
    )

    add_option(
        "-u",
        "--units",
        "units_",
        multiple=True,
        default=None,
        type=click.Choice(["M", "mM", "uM", "nM", "pM", "noconv"]),
        help=(
            "Unit of the activity given. Need to supply either"
            " None or as many as property_columns."
        ),
    )

    add_option(
        "-v",
        "--verbose",
        "verbose_",
        count=True,
        default=0,
        help=(
            "use -v to set logging level to info, use -vv to set logging level"
            " to debug. without -v logging level is set to warning."
        ),
    )

    add_option(
        "--log",
        "log_file_",
        default=None,
        type=str,
        help=(
            "File to write all logging info to. "
            "If none is provided, logging is displayed in commandline. "
            "It will be placed in the output directory."
        ),
    )

    def make_input_options_wrapper(**kwargs) -> Callable:
        """Create input options instance.

        creates an input options object from the flags provided in the command
        line and parses them so the original command is only taking the
        inputoptions object but not every single command line flag.

        Returns:
            Callable: command with input options as argument.
        """
        in_kwargs = {}
        for name in param_names:
            value = kwargs.pop(name)
            in_kwargs[name] = value

        kwargs["input_options"] = InputOptions(**in_kwargs)
        return command(**kwargs)

    set_click_attrs(make_input_options_wrapper, command)
    return make_input_options_wrapper


class InputOptions:
    """Handles the type checking for the."""

    def __init__(
        self,
        infile_: Path,
        update_: bool,
        max_heavy_: int,
        max_heavy_in_transformation_: int,
        classify_: bool,
        canonicalize_: bool,
        no_chiral_: bool,
        delimiter_: str,
        include_censored_: bool,
        directory_: Path | None,
        series_column_: str | None,
        property_columns_: tuple[str, ...],
        units_: tuple[str | None, ...],
        verbose_: int,
        log_file_: str | None,
    ) -> None:
        """Create input Option instance with cli arguments.

        Args:
            infile_ (Path): Path to input file
            update_ (bool): whether to use old fragmentation
            max_heavy_ (int): max heavy atoms per ligand
            max_heavy_in_transformation_ (int): max heavy input in transformation
            classify_ (bool): whether to run classification
            canonicalize_ (bool): wheter to provide canonicalized output
            no_chiral_ (bool): whether to skip chiral transformation
            delimiter_ (str): delimiter used in input file
            include_censored_ (bool): whether to include censored values
            directory_ (Path | None): where to store intermediate and output files
            series_column_ (str | None): column containing info about subseries
            property_columns_ (tuple[str, ...]): name of the columns with properties
            to analyze
            units_ (tuple[str  |  None, ...]): units of provided proeprty columns
            verbose_ (int): whether to give verbose output
            log_file_ (str | None): file to log log messages to

        Raises:
            ValueError: Raised if any input argument is not valid.
        """
        self.infile = infile_
        self.update = update_
        self.max_heavy = max_heavy_
        self.classify = classify_
        self.canonicalize = canonicalize_
        self.max_heavy_in_transformation = max_heavy_in_transformation_
        self.no_chiral = no_chiral_
        self.delimiter = delimiter_
        self.include_censored = include_censored_
        self.directory = directory_
        self.series_column = series_column_
        self.property_columns = property_columns_
        self.units = units_
        self.verbose = verbose_
        self.log_file = log_file_
        if not self._property_columns_and_units_correct():
            raise ValueError(
                (
                    "If Units is provided, the number of units has to be the "
                    "same as the number of property_columns provided."
                ),
            )

    def _property_columns_and_units_correct(self) -> bool:
        return len(self.property_columns) == len(self.units)

    @property
    def infile(self) -> Path:
        """Path to input file for analysis."""
        return self._infile

    @infile.setter
    def infile(self, value: Path) -> None:
        self._infile = value

    @property
    def update(self) -> bool:
        """Whether to use previous fragmentation."""
        return self._update

    @update.setter
    def update(self, value: bool) -> None:
        self._update = value

    @property
    def max_heavy(self) -> int:
        """Max num heavy atoms in ligands."""
        return self._max_heavy

    @max_heavy.setter
    def max_heavy(self, value: int) -> None:
        self._max_heavy = value

    @property
    def classify(self) -> bool:
        """Whether to classify output."""
        return self._classify

    @classify.setter
    def classify(self, value: bool) -> None:
        self._classify = value

    @property
    def canonicalize(self) -> bool:
        """Whether to canonicalize output."""
        return self._canonicalize

    @canonicalize.setter
    def canonicalize(self, value: bool) -> None:
        self._canonicalize = value

    @property
    def max_heavy_in_transformation(self) -> int:
        """Max num heavy atoms in transformations."""
        return self._max_heavy_in_transformation

    @max_heavy_in_transformation.setter
    def max_heavy_in_transformation(self, value: int) -> None:
        self._max_heavy_in_transformation = value

    @property
    def no_chiral(self) -> bool:
        """Whether to use chiral transformations in analysis."""
        return self._no_chiral

    @no_chiral.setter
    def no_chiral(self, value: bool) -> None:
        self._no_chiral = value

    @property
    def delimiter(self) -> str:
        """Delimiter used in the input file."""
        return self._delimiter

    @delimiter.setter
    def delimiter(self, value: str) -> None:
        self._delimiter = value

    @property
    def include_censored(self) -> bool:
        """Whether to include censored values."""
        return self._include_censored

    @include_censored.setter
    def include_censored(self, value: bool) -> None:
        self._include_censored = value

    @property
    def directory(self) -> Path:
        """Directory for where to store output."""
        return self._directory

    @directory.setter
    def directory(self, value: Path | None) -> None:
        if value is None:
            value = Path(os.path.dirname(self.infile)).resolve()
        if not value.exists():
            os.makedirs(value)
        self._directory = value

    @property
    def series_column(self) -> str | None:
        """Name of series column if any."""
        return self._series_column

    @series_column.setter
    def series_column(self, value: str | None) -> None:
        self._series_column = value

    @property
    def property_columns(self) -> list[str]:
        """List of columns containig to analyse properties."""
        return self._property_columns

    @property_columns.setter
    def property_columns(self, value: tuple[str, ...]) -> None:
        self._property_columns = list(value)

    @property
    def units(self) -> list[str | None]:
        """List of units."""
        return self._units  # type: ignore

    @units.setter
    def units(self, value: tuple[str | None, ...]) -> None:
        if len(value) == 0:
            self._units = [None for _ in self.property_columns]
            return
        self._units = list(value)

    @property
    def verbose(self) -> int:
        """Log level."""
        return self._verbose

    @verbose.setter
    def verbose(self, value: int) -> None:
        try:
            value = int(value)
        except (ValueError, TypeError) as exc:
            raise ValueError(
                (
                    "Verbosity flag is used wrong, please use 'nonadditivity"
                    " --help' to get more info"
                ),
            ) from exc
        value = min(value, 2)
        value = max(value, 0)
        self._verbose = VERBOSITY_LEVEL[value]

    @property
    def log_file(self) -> Path | None:
        """Log file path for nonadditivity analysis."""
        return self._log_file

    @log_file.setter
    def log_file(self, value: str | None) -> None:
        if value is None:
            self._log_file = None
            return
        self._log_file = self.directory / value



"""Implement Transformation class for classification.

Holds Class for Describing Properties on a Transformation Level.
This is used for the NonadditivityAnalysis Package.
"""

from __future__ import annotations

from enum import Enum, auto
from typing import TYPE_CHECKING, ClassVar

from nonadditivity.classification.ortho_classification import (
    ortho_substituent_exchanged,
    ortho_substituent_introduced,
)
from nonadditivity.classification.rgroup_distance import (
    get_constant_and_transformation_indices,
    get_r_group_anchor,
)
from nonadditivity.classification.transformation_classification import (
    calculate_fp_similarity,
    get_num_cuts,
    get_num_heavy_atoms_rgroups,
    is_h_replaced,
    stereo_classify_transformation,
    tertiary_amide_generated,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from nonadditivity.classification.classification_classes.compound import Compound


class Transformation:
    """Class holding transformation classification.

    Transformation class holds two Compound object as well as strings
    for the constant part as well as the transformation part of the
    Transformation. It also calculates the properties enumerated in
    Properties.
    """

    class Properties(Enum):
        """Enumerates type of compound classifications."""

        IS_H_REPLACED = auto()
        NUM_HEAVY_ATOMS_IN_RGROUPS = auto()
        ORTHO_SUBSTITUENT_CHANGES = auto()
        ORTHO_SUBSTITUENT_INTRODUCED = auto()
        TERTIARY_AMIDE_FORMED = auto()
        TRANSFORMATION_STEREO = auto()
        MFP2_SIMILARITY = auto()
        NUM_MMPDB_CUTS = auto()

    classification_function: ClassVar[dict[Properties, Callable]] = {
        Properties.IS_H_REPLACED: is_h_replaced,
        Properties.NUM_HEAVY_ATOMS_IN_RGROUPS: get_num_heavy_atoms_rgroups,
        Properties.ORTHO_SUBSTITUENT_CHANGES: ortho_substituent_exchanged,
        Properties.ORTHO_SUBSTITUENT_INTRODUCED: ortho_substituent_introduced,
        Properties.TERTIARY_AMIDE_FORMED: tertiary_amide_generated,
        Properties.TRANSFORMATION_STEREO: stereo_classify_transformation,
        Properties.MFP2_SIMILARITY: calculate_fp_similarity,
        Properties.NUM_MMPDB_CUTS: get_num_cuts,
    }

    classification_keys: ClassVar[dict[Properties, str]] = {
        Properties.IS_H_REPLACED: "h_replaced",
        Properties.NUM_HEAVY_ATOMS_IN_RGROUPS: "num_heavy_atoms_rgroups",
        Properties.ORTHO_SUBSTITUENT_CHANGES: "ortho_substituent_exchanged",
        Properties.ORTHO_SUBSTITUENT_INTRODUCED: "ortho_substituent_introduced",
        Properties.TRANSFORMATION_STEREO: "transformation_stere_classification",
        Properties.TERTIARY_AMIDE_FORMED: "tertiary_amide_formed",
        Properties.MFP2_SIMILARITY: "mfp_2_similarity",
        Properties.NUM_MMPDB_CUTS: "num_mmpdb_cuts",
    }

    def __init__(
        self,
        compound_1: Compound,
        compound_2: Compound,
        constant_smarts: str,
        transformation_smarts: str,
    ) -> None:
        """Creat transformation instance.

        creates Transformation object holding 2 compounds
        as well as smarts for constant and transformation part.
        Upon instanciation, the indices for the constant and
        transformation part are calculatd for the rdkit molecules.

        Args:
            compound_1 (Compound): compound 1
            compound_2 (Compound): compound 2
            constant_smarts (str): constant smarts pattern
            transformation_smarts (str): transformation smarts pattern
        """
        self.compound_1 = compound_1
        self.compound_2 = compound_2
        self.constant_smarts = constant_smarts
        self.transformation_smarts = transformation_smarts.split(">>")

        (
            self.compound_1_constant_indices,
            self.compound_1_transformation_indices,
        ) = get_constant_and_transformation_indices(
            transformation_smarts=self.transformation_smarts[0],
            constant_smarts=constant_smarts,
            molecule=compound_1.rdkit_molecule,
        )

        (
            self.compound_2_constant_indices,
            self.compound_2_transformation_indices,
        ) = get_constant_and_transformation_indices(
            transformation_smarts=self.transformation_smarts[1],
            constant_smarts=constant_smarts,
            molecule=compound_2.rdkit_molecule,
        )

        self.compound_1_anchor = get_r_group_anchor(
            constant_indices=self.compound_1_constant_indices,
            transformation_indices=self.compound_1_transformation_indices,
        )
        self.compound_2_anchor = get_r_group_anchor(
            constant_indices=self.compound_2_constant_indices,
            transformation_indices=self.compound_2_transformation_indices,
        )
        self._classification: dict[
            Transformation.Properties,
            bool | int | list[int] | float,
        ] = {}

    def get_property(
        self,
        key: Properties,
    ) -> bool | int | list[int] | float:
        """Calculate property.

        Get a Property of a Transformation, all in Transformation.Properties are
        all-owed.

        Args:
            key (CompoundProperties): property name

        Returns:
            Union[bool,int]: Transformation porperty
        """
        try:
            return self._classification[key]
        except KeyError:
            self._classification[key] = Transformation.classification_function[key](
                compounds=[self.compound_1, self.compound_2],
                transformation_smarts=self.transformation_smarts,
                constant_smarts=self.constant_smarts,
                constant_indices=[
                    self.compound_1_constant_indices,
                    self.compound_2_constant_indices,
                ],
                anchors=(self.compound_1_anchor, self.compound_2_anchor),
            )
            return self.get_property(key=key)

    def classify(self) -> None:
        """Calculate all properties enumerated in Transformation.Properties."""
        for prop in Transformation.Properties:
            if prop not in self._classification:
                self.get_property(prop)

    def get_classification_dict(
        self,
        force_classification: bool = True,
    ) -> dict[Properties, bool | int | list[int] | float]:
        """Get dictionary with already calculated properties.

        Args:
            force_classification (bool, optional): Forces calculation of all properties.
                Defaults to True.

        Returns:
            dict[str, int]: calculated properties
        """
        if force_classification:
            self.classify()
        return self._classification

    def get_formatted_classification_dict(
        self,
        force_classification: bool = True,
    ) -> dict[str, bool | int | list[int] | float]:
        """Get dictionary with strings for dataframe and calcualted properties.

        Return a dictionary with the properties already
        calculated for the compound with strings as keys
        instead of the Compound.Properties

        Args:
            force_classification (bool, optional): classifies the
            whole transformation. Defaults to True.

        Returns:
            dict[str, int]: calculated properties 'nicely' formatted.
        """
        if force_classification:
            self.classify()
        return {
            Transformation.classification_keys[key]: value
            for key, value in self._classification.items()
        }

    def get_reverse(self) -> Transformation:
        """Get reversed transformation object.

        Returns a Transformation object, that desctibes the
        transformation that is the reverse of this objects
        transforamtion.

        Returns:
            Transformation: reverse transformation.
        """
        transformation_smart = (
            f"{self.transformation_smarts[1]}>>{self.transformation_smarts[0]}"
        )
        return Transformation(
            compound_1=self.compound_2,
            compound_2=self.compound_1,
            constant_smarts=self.constant_smarts,
            transformation_smarts=transformation_smart,
        )



"""Implement Circle class for classification.

Holds Class for Describing Properties on a circle Level.
This is used for the NonadditivityAnalysis Package.
"""

from collections.abc import Callable
from enum import Enum, auto
from typing import ClassVar

from nonadditivity.classification.circle_classification import (
    circle_has_stereoinversion_in_transformation,
    get_chi0_diff,
    get_chi1_diff,
    get_chi2_diff,
    get_compound_stereocenter_classification,
    get_formal_charge_diff,
    get_hbond_acceptor_diff,
    get_hbond_donor_diff,
    get_log_p_diff,
    get_max_num_heavy_atom_in_transformation,
    get_max_num_mmpdb_cuts,
    get_min_transformation_tanimoto,
    get_num_heavy_atoms_diff,
    get_num_rot_bonds_diff,
    get_sp3_carbon_diff,
    get_tertiary_amide_formed,
    get_tpsa_diff,
    substituents_in_same_ring_system,
)
from nonadditivity.classification.classification_classes.compound import Compound
from nonadditivity.classification.classification_classes.transfromation import (
    Transformation,
)
from nonadditivity.classification.ortho_classification import (
    get_ortho_transformation_changing,
    get_ortho_transformation_introduced,
)
from nonadditivity.classification.rgroup_distance import get_num_atoms_between_rgroups


class Circle:
    """Object for classifying circles.

    The Circle Class holds all relevant information for classifying a NAA double
    transformation cycle. It holds the four transformation objects of a circle
    and hence also the four compounds the circle is made from.
    """

    class Properties(Enum):
        """Enumerates type of compound classifications."""

        DISTANCE_BETWEEN_R_GROUPS = auto()
        H_BOND_DONORS_DIFF = auto()
        H_BOND_ACCEPTORS_DIFF = auto()
        FORMAL_CHARGE_DIFF = auto()
        POLAR_SURFACE_AREA_DIFF = auto()
        NUM_ROTATABLE_BONDS_DIFF = auto()
        SP3_CARBON_DIFF = auto()
        LOG_P_DIFF = auto()
        CHI0_DIFF = auto()
        CHI1_DIFF = auto()
        CHI2_DIFF = auto()
        NUM_HEAVY_ATOMS_DIFF = auto()
        HAS_TRANSFORMATION_AT_ORTHO = auto()
        HAS_ORTHO_SUBSTITUENT_INTRODUCED = auto()
        HAS_TERTIARY_AMIDE_FORMED = auto()
        HAS_INVERSION_IN_TRANSFORMATION = auto()
        MAX_NUM_MMPDB_CUTS = auto()
        MAX_HEAVY_ATOM_IN_TRANSFORMATION = auto()
        COMPOUND_STEREO_CLASSIFICATION = auto()
        MIN_TANIMOTO = auto()
        SUBSTITUENT_ON_SAME_RING_SYSYTEM = auto()

    classification_function: ClassVar[dict[Properties, Callable]] = {
        Properties.DISTANCE_BETWEEN_R_GROUPS: get_num_atoms_between_rgroups,
        Properties.H_BOND_DONORS_DIFF: get_hbond_donor_diff,
        Properties.H_BOND_ACCEPTORS_DIFF: get_hbond_acceptor_diff,
        Properties.FORMAL_CHARGE_DIFF: get_formal_charge_diff,
        Properties.POLAR_SURFACE_AREA_DIFF: get_tpsa_diff,
        Properties.NUM_ROTATABLE_BONDS_DIFF: get_num_rot_bonds_diff,
        Properties.SP3_CARBON_DIFF: get_sp3_carbon_diff,
        Properties.LOG_P_DIFF: get_log_p_diff,
        Properties.NUM_HEAVY_ATOMS_DIFF: get_num_heavy_atoms_diff,
        Properties.CHI0_DIFF: get_chi0_diff,
        Properties.CHI1_DIFF: get_chi1_diff,
        Properties.CHI2_DIFF: get_chi2_diff,
        Properties.HAS_TRANSFORMATION_AT_ORTHO: get_ortho_transformation_changing,
        Properties.HAS_ORTHO_SUBSTITUENT_INTRODUCED: get_ortho_transformation_introduced,
        Properties.HAS_TERTIARY_AMIDE_FORMED: get_tertiary_amide_formed,
        Properties.HAS_INVERSION_IN_TRANSFORMATION: circle_has_stereoinversion_in_transformation,
        Properties.MAX_NUM_MMPDB_CUTS: get_max_num_mmpdb_cuts,
        Properties.MAX_HEAVY_ATOM_IN_TRANSFORMATION: get_max_num_heavy_atom_in_transformation,
        Properties.COMPOUND_STEREO_CLASSIFICATION: get_compound_stereocenter_classification,
        Properties.MIN_TANIMOTO: get_min_transformation_tanimoto,
        Properties.SUBSTITUENT_ON_SAME_RING_SYSYTEM: substituents_in_same_ring_system,
    }
    classification_keys: ClassVar[dict[Properties, str]] = {
        Properties.DISTANCE_BETWEEN_R_GROUPS: "num_atoms_between_r_groups",
        Properties.H_BOND_DONORS_DIFF: "hbond_donor_diff",
        Properties.H_BOND_ACCEPTORS_DIFF: "hbond_acceptor_diff",
        Properties.FORMAL_CHARGE_DIFF: "formal_charge_diff",
        Properties.POLAR_SURFACE_AREA_DIFF: "tpsa_diff",
        Properties.NUM_ROTATABLE_BONDS_DIFF: "num_rot_bonds_diff",
        Properties.SP3_CARBON_DIFF: "sp3_carbon_diff",
        Properties.LOG_P_DIFF: "log_p_diff",
        Properties.CHI0_DIFF: "chi0_diff",
        Properties.CHI1_DIFF: "chi1_diff",
        Properties.CHI2_DIFF: "chi2_diff",
        Properties.NUM_HEAVY_ATOMS_DIFF: "num_heavy_atoms_diff",
        Properties.HAS_TRANSFORMATION_AT_ORTHO: "transformation_at_ortho",
        Properties.HAS_ORTHO_SUBSTITUENT_INTRODUCED: "ortho_substituent_introduced",
        Properties.HAS_TERTIARY_AMIDE_FORMED: "tertiary_amide_formed",
        Properties.HAS_INVERSION_IN_TRANSFORMATION: "has_stereoinversion_in_transformation",
        Properties.MAX_NUM_MMPDB_CUTS: "max_num_mmpdb_cuts",
        Properties.MAX_HEAVY_ATOM_IN_TRANSFORMATION: "max_num_heavy_atom_in_transformation",
        Properties.COMPOUND_STEREO_CLASSIFICATION: "compound_stereocenter_classification",
        Properties.MIN_TANIMOTO: "min_transformation_tanimoto",
        Properties.SUBSTITUENT_ON_SAME_RING_SYSYTEM: "substituents_in_same_ring_system",
    }

    def __init__(
        self,
        transformation_1: Transformation,
        transformation_2: Transformation,
        transformation_3: Transformation,
        transformation_4: Transformation,
    ) -> None:
        """Instanciate circle classification object.

        Args:
            transformation_1 (Transformation): transformation 1 in the circle
            transformation_2 (Transformation): transformation 2 in the circle
            transformation_3 (Transformation): transformation 3 in the circle
            transformation_4 (Transformation): transformation 4 in the circle
        """
        self.transformation_1 = transformation_1
        self.transformation_2 = transformation_2
        self.transformation_3 = transformation_3
        self.transformation_4 = transformation_4
        self._classification: dict[Circle.Properties, int] = {}

    def compounds(self) -> list[Compound]:
        """Get compounds in the circle.

        Returns list of compounds as they are named in the get_circles function in
        nonadditivty.nonadditivty_core

        Returns:
            list[Compound]: list of compounds in circle
        """
        return [
            self.transformation_1.compound_1,
            self.transformation_1.compound_2,
            self.transformation_4.compound_2,
            self.transformation_4.compound_1,
        ]

    def get_property(self, key: Properties) -> int | tuple[int] | bool:
        """Calculate a Property of a compound, all in Circle.Properties are allowed.

        Args:
            key (CompoundProperties): property name

        Returns:
            int: compound porperty
        """
        try:
            return self._classification[key]
        except KeyError:
            self._classification[key] = Circle.classification_function[key](
                transformation_1=self.transformation_1,
                transformation_2=self.transformation_2,
                transformation_3=self.transformation_3,
                transformation_4=self.transformation_4,
            )
            return self.get_property(key=key)

    def classify(self) -> None:
        """Calculate all properties enumerated in Circle.Properties."""
        for prop in Circle.Properties:
            if prop not in self._classification:
                _ = self.get_property(prop)

    def get_classification_dict(
        self,
        force_classification: bool = True,
    ) -> dict[Properties, int]:
        """Get dictionary with the properties already calculated for the Circle.

        Args:
            force_classification (bool, optional): Forces calculation of all properties.
            Defaults to True.

        Returns:
            dict[str, int]: calculated properties
        """
        if force_classification:
            self.classify()
        return self._classification

    def get_formatted_classification_dict(
        self,
        force_classification: bool = True,
    ) -> dict[str, int]:
        """Get dictionary with strings for dataframe and calcualted properties.

        Return a dictionary with the properties already calculated for the compound
        with strings as keys instead of the Compound.Properties

        Args:
            force_classification (bool, optional): whether to force all possible
            classification. Defaults to True.

        Returns:
            dict[str, int]: formatted classificaton dictionary.
        """
        if force_classification:
            self.classify()
        return {
            Circle.classification_keys[key]: value
            for key, value in self._classification.items()
        }



"""Implements transformation classification.

All these functions are higly specific, do not use them outside
of their intended application.
"""

import contextlib
from collections.abc import Callable
from typing import TYPE_CHECKING

from rdkit import Chem, DataStructs

from nonadditivity.classification.utils import is_unique
from nonadditivity.utils.types import Atom, Molecule

if TYPE_CHECKING:
    from nonadditivity.classification import Compound

AMIDE_PATTERN = Chem.MolFromSmarts(  # type:ignore pylint:disable=E1101
    "[NX3][CX3](=[OX1])[#6]",
)


def get_num_cuts(**kwargs: str) -> int:
    """Returns number of mmpdb cuts made in a transformation.

    use keyword argument 'transformation_smarts' for this function to work.


    Returns:
        int: number of mmpdb cuts
    """
    smarts: str = kwargs["transformation_smarts"][0]
    num = 0
    for i in range(3):
        if f"[*:{i+1!s}]" in smarts:
            num += 1
    return num


def is_h_replaced(**kwargs: str) -> bool:
    """Check whether one side of the transformation is only a hydrogen.

    Returns true if an H is replaced or something is replaced by
    a H in this transformation.

    use keyword argumnt 'transformation_smarts' for it to work.

    Returns:
        bool: True if H is part of transformation
    """
    return "[*:1][H]" in kwargs["transformation_smarts"]


def _tertiary_nitrogen_generated(**kwargs) -> int:
    """Check whether a tertiary nitrogen is formed in transformation.

    use keyword arguments 'compounds' and 'anchors' for this function to work.

    Returns:
        int: 0 == No, 1 == Yes, -1 == Not identifiable, 2 == yes and in both cases
        the nitrogen is conjugated..
    """

    def is_conjugated(atom: Atom) -> bool:
        return any(b.GetIsConjugated() for b in atom.GetBonds())

    if not is_h_replaced(**kwargs):
        return 0
    molecule_1, molecule_2 = (cpd.rdkit_molecule for cpd in kwargs["compounds"])
    anchor_1_idxs, anchor_2_idxs = kwargs["anchors"]
    for val in (-1, 0):
        if any(is_unique(a) == val for a in (anchor_1_idxs, anchor_2_idxs)):
            return val
    anchor_1_atom, anchor_2_atom = (
        molecule_1.GetAtomWithIdx(
            anchor_1_idxs[0],
        ),
        molecule_2.GetAtomWithIdx(
            anchor_2_idxs[0],
        ),
    )
    if (
        any(a.GetIsAromatic() for a in (anchor_1_atom, anchor_2_atom))
        or anchor_1_atom.GetSymbol() == anchor_2_atom.GetSymbol() != "N"
        or {a.GetTotalNumHs() for a in [anchor_1_atom, anchor_2_atom]} != {0, 1}
    ):
        return 0
    if any(is_conjugated(a) for a in (anchor_1_atom, anchor_2_atom)):
        return 2
    return 1


def tertiary_amide_generated(**kwargs) -> int:
    """Check Whether tertiary amide is formed in transformation.

    use keyword arguments 'compounds' and 'anchors' for this function to work.

    Returns:
        int: 0 == No | Not identifiable, 1 == Yes.
    """

    def anchor_is_in_pattern(pattern: Molecule, **kwargs) -> bool:
        molecule = kwargs["compounds"][0].rdkit_molecule
        anchor = kwargs["anchors"][0][0]
        potential_matches = molecule.GetSubstructMatches(pattern)
        return any(anchor in m for m in potential_matches)

    tertiary_nitrogen = _tertiary_nitrogen_generated(
        **kwargs,
    )
    if tertiary_nitrogen < 2:
        return 0
    return int(anchor_is_in_pattern(pattern=AMIDE_PATTERN, **kwargs))


def stereo_classify_transformation(**kwargs) -> int:
    """Check whether transforamtion is only changing stereochemistry.

    Checks whether a transformation is only a stereoinversion or going
    from a compound with assigned stereochemistry to the same compound
    without assigned stereocenters.

    use keyword arguments 'compounds'
    for this function to work.

    Returns:
        int: 0 == Transformation not stereo only
            1 == assigned -> unassigned
                2 == stereoinversion
    """
    molecules: list[Molecule] = [c.rdkit_molecule for c in kwargs["compounds"]]

    # if not same number of atoms, change has to be other than stereochemistry!
    if molecules[0].GetNumAtoms() != molecules[1].GetNumAtoms():
        return 0
    # if we can match without stereochemistry, the transformation has to be
    # stereo related
    if (molecules[0].GetSubstructMatch(molecules[1], useChirality=False)) and (
        molecules[1].GetSubstructMatch(molecules[0], useChirality=False)
    ):
        # Unassigned stereocenters are matched with assigned stereocenters
        # --> we go from assigned to unassigned or vice versa
        if molecules[0].GetSubstructMatch(molecules[1], useChirality=True):
            return 1
        # compounds are matched without chirality but not with chirality,
        # --> change is only stereochem related!
        return 2
    # if we cannot match the two hole structures, transformation is not
    # stereochemistry related.
    return 0


def _get_phys_chem_property(
    function: Callable[[Molecule], float],
    **kwargs,
) -> list[float]:
    """Apply function to both transformation parts.

    Takes the transformation smarts, converts them to smiles,
    and then lets the function caluclate a value for the molecules
    from the just generated smiles.

    use keyword argument 'compounds' and 'transformation_smarts'
    for this function to work.

    Args:
        function (Callable[[Molecule], float]): rdkit description function.
        add_hs (bool, optional): Whether to add Hydrogens. Defaults to False.
        **kwargs: kwargs used for this function.

    Returns:
        list[float]: values calculated.
    """
    mols = [
        Chem.MolFromSmarts(c)  # type: ignore pylint: disable=E1101
        for c in kwargs["transformation_smarts"]
    ]
    for mol in mols:
        with contextlib.suppress(Exception):
            Chem.SanitizeMol(mol)  # type: ignore pylint: disable=E1101
    return [function(mol) for mol in mols]


def get_num_heavy_atoms_rgroups(**kwargs: str) -> list[int]:
    """Calculate number of heavy atoms in the rgroups of a transformation.

    use keyword argument 'transformation_smarts' for this function to work.

    Returns:
        int: difference in heavy atoms
    """

    def get_num_heavy(molecule: Molecule) -> int:
        return molecule.GetNumHeavyAtoms()

    return [int(x) for x in _get_phys_chem_property(function=get_num_heavy, **kwargs)]


def calculate_fp_similarity(**kwargs) -> float:
    """Calculate Tanimoto of Morgan fingerprints (2) of compounds in transformation.

    use keyword argument 'compounds' for this function to work.

    Returns:
        float: Tanimoto of Fingerprints
    """
    compounds: list[Compound] = kwargs["compounds"]
    return DataStructs.TanimotoSimilarity(  # type: ignore pylint: disable=E1101
        *[m.get_property(m.Properties.MORGANFP) for m in compounds],
    )



"""Utilitary functions for the classification workflow."""

from collections.abc import Sequence

from rdkit import Chem


def list_empty(nested_list: Sequence) -> bool:
    """Check whether a list or nested lists are empty.

    Args:
        nested_list (list): list to check.

    Returns:
        bool: True when all nested lists are empty
    """
    if isinstance(nested_list, list):
        return all(map(list_empty, nested_list))
    return False


def is_unique(anchor: int | Sequence[int | Sequence[int]]) -> int:
    """Check whether an anchor is unique.

    Args:
        anchor (int | Sequence[int | Sequence[int]]): list of atom indices.

    Returns:
        int: 1 if unique and length 1, 0 if unique and length > 1, -1 if not unique.
    """
    if not anchor or list_empty(anchor):  # type: ignore
        return -1
    assert isinstance(anchor, Sequence)
    if not isinstance(anchor[0], int):
        return -1
    return int(len(anchor) == 1)


def is_match_unique(
    match: int | Sequence[int | Sequence[int]],
) -> bool:
    """Return True if the list is not nested and not empty.

    Args:
        match (int | Sequence[int | Sequence[int]]): list to check.

    Returns:
        bool: True if list not nested and not empty.
    """
    if isinstance(match, int):
        return True
    if list_empty(match):
        return False
    return isinstance(match[0], int)


def convert_smarts_to_smiles(smarts: str) -> str:
    """Convert smarts to smiles.

    Crudely converts SMARTS to SMILES by creating a rdkit Molecule
    object from the SMARTS and writing it to SMILES.

    Args:
        smarts (str): input SMARTS

    Returns:
        str: SMILES
    """
    try:
        return Chem.MolToSmiles(  # type: ignore pylint: disable=E1101
            Chem.MolFromSmarts(smarts),  # type: ignore pylint: disable=E1101
        )
    except Exception as exc:
        raise TypeError from exc
